"""
Utility functions and helpers for the Observicia SDK.
"""

from .helpers import (get_current_span, get_current_context, count_tokens,
                      format_trace_attributes, safe_json_serialize)

__all__ = [
    "get_current_span", "get_current_context", "count_tokens",
    "format_trace_attributes", "safe_json_serialize"
]



from functools import wraps
from typing import Any, Callable, Dict, List, Optional
import asyncio
from opentelemetry import trace
from opentelemetry.trace import SpanKind
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, ConsoleSpanExporter
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter

from .context_manager import ObservabilityContext
from .policy_engine import PolicyEngine, PolicyResult
from .token_tracker import TokenTracker
from ..utils.tracing_helpers import start_llm_span
from ..utils.policy_helpers import enforce_policies


class TracingClient:
    """Client for handling traces with policy enforcement."""

    def __init__(self,
                 service_name: str,
                 policy_engine: Optional[PolicyEngine] = None,
                 token_tracker: Optional[TokenTracker] = None,
                 otel_endpoint: Optional[str] = None,
                 trace_console: bool = False):
        """Initialize the tracing client."""
        self.service_name = service_name
        self.policy_engine = policy_engine
        self.token_tracker = token_tracker
        self.tracer = trace.get_tracer(service_name)

        # Set up trace export
        provider = TracerProvider()
        if otel_endpoint:
            otlp_processor = BatchSpanProcessor(
                OTLPSpanExporter(endpoint=otel_endpoint))
            provider.add_span_processor(otlp_processor)
        if trace_console:
            console_processor = BatchSpanProcessor(ConsoleSpanExporter())
            provider.add_span_processor(console_processor)
        trace.set_tracer_provider(provider)

    def start_span(
        self,
        name: str,
        context: Optional[Dict[str, Any]] = None,
        kind: SpanKind = SpanKind.CLIENT,
        attributes: Optional[Dict[str, Any]] = None,
    ) -> trace.Span:
        """Start a new span with context and attributes."""
        return start_llm_span(
            name, {
                **(attributes or {}), "service.name": self.service_name
            })


class TraceDecorator:
    """Decorator for adding tracing and policy enforcement to functions."""

    def __init__(self,
                 tracing_client: TracingClient,
                 policies: Optional[List[str]] = None,
                 trace_level: str = "normal",
                 enforce_policies: bool = True):
        self.client = tracing_client
        self.policies = policies or []
        self.trace_level = trace_level
        self.enforce_policies = enforce_policies

    def __call__(self, func: Callable) -> Callable:
        """Apply the decorator to a function."""

        @wraps(func)
        async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
            with self.client.start_span(
                    name=func.__name__,
                    context={
                        "args": args,
                        "kwargs": kwargs
                    },
                    attributes={"trace.level": self.trace_level}) as span:
                try:
                    if self.enforce_policies:
                        await enforce_policies(self.client.policy_engine, span,
                                               {
                                                   "args": args,
                                                   "kwargs": kwargs,
                                                   "phase": "input"
                                               })

                    result = await func(*args, **kwargs)

                    if self.enforce_policies:
                        await enforce_policies(self.client.policy_engine, span,
                                               {
                                                   "result": result,
                                                   "phase": "output"
                                               })

                    return result

                except Exception as e:
                    span.record_exception(e)
                    raise

        @wraps(func)
        def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
            with self.client.start_span(
                    name=func.__name__,
                    context={
                        "args": args,
                        "kwargs": kwargs
                    },
                    attributes={"trace.level": self.trace_level}) as span:
                try:
                    if self.enforce_policies:
                        enforce_policies(self.client.policy_engine, span, {
                            "args": args,
                            "kwargs": kwargs,
                            "phase": "input"
                        })

                    result = func(*args, **kwargs)

                    if self.enforce_policies:
                        enforce_policies(self.client.policy_engine, span, {
                            "result": result,
                            "phase": "output"
                        })

                    return result

                except Exception as e:
                    span.record_exception(e)
                    raise

        return async_wrapper if asyncio.iscoroutinefunction(
            func) else sync_wrapper


# Convenience decorators
def trace(policies: Optional[List[str]] = None,
          trace_level: str = "normal",
          enforce_policies: bool = True):
    """Decorator for basic tracing and policy enforcement."""
    context = ObservabilityContext.get_current()
    return TraceDecorator(context.tracing_client,
                          policies=policies,
                          trace_level=trace_level,
                          enforce_policies=enforce_policies)


def trace_rag(policies: Optional[List[str]] = None,
              trace_level: str = "detailed"):
    """Decorator specifically for RAG applications."""
    context = ObservabilityContext.get_current()
    return TraceDecorator(context.tracing_client,
                          policies=policies
                          or ["retrieval_relevance", "context_usage"],
                          trace_level=trace_level)


def trace_stream(policies: Optional[List[str]] = None,
                 trace_level: str = "normal"):
    """Decorator for streaming responses."""
    context = ObservabilityContext.get_current()
    return TraceDecorator(context.tracing_client,
                          policies=policies,
                          trace_level=trace_level)



# token_helpers.py
"""Utility functions for token counting and tracking"""

import tiktoken
from typing import List, Dict, Any
from opentelemetry.trace import Span
from ..core.token_tracker import TokenTracker


def count_prompt_tokens(messages: List[Dict[str, Any]], model: str) -> int:
    """Count tokens in chat messages."""
    try:
        encoding = tiktoken.encoding_for_model(model)
        num_tokens = 0
        for message in messages:
            if isinstance(message.get('content'), str):
                num_tokens += len(encoding.encode(message['content']))
            num_tokens += 4  # Format tokens per message
        num_tokens += 2  # Conversation format tokens
        return num_tokens
    except Exception:
        # Fallback to approximate count
        return sum(
            len(str(msg.get('content', '')).split()) for msg in messages)


def count_text_tokens(text: str, model: str) -> int:
    """Count tokens in plain text."""
    try:
        encoding = tiktoken.encoding_for_model(model)
        return len(encoding.encode(text))
    except Exception:
        # Fallback to approximate count
        return len(text.split())


def record_token_usage(span: Span, response: Any) -> None:
    """Record token usage in span from response."""
    if hasattr(response, 'usage'):
        usage = response.usage
        usage_dict = {}

        # Handle different usage structures
        if hasattr(usage, 'prompt_tokens'):
            usage_dict['prompt.tokens'] = usage.prompt_tokens
        if hasattr(usage, 'completion_tokens'):
            usage_dict['completion.tokens'] = usage.completion_tokens
        elif hasattr(usage, 'total_tokens') and hasattr(
                usage, 'prompt_tokens'):
            # Some APIs only give total and prompt, derive completion
            usage_dict[
                'completion.tokens'] = usage.total_tokens - usage.prompt_tokens
        if hasattr(usage, 'total_tokens'):
            usage_dict['total.tokens'] = usage.total_tokens

        # Set all available usage metrics
        span.set_attributes(usage_dict)


def update_token_usage(token_tracker: TokenTracker, provider: str,
                       usage: Any) -> None:
    """Update token tracker with usage statistics."""
    prompt_tokens = getattr(usage, 'prompt_tokens', 0)
    completion_tokens = getattr(usage, 'completion_tokens', 0)

    # If we have total but not completion, derive it
    if hasattr(usage,
               'total_tokens') and not hasattr(usage, 'completion_tokens'):
        completion_tokens = usage.total_tokens - prompt_tokens

    token_tracker.update(provider,
                         prompt_tokens=prompt_tokens,
                         completion_tokens=completion_tokens)



import asyncio
from functools import wraps
from typing import Any, AsyncIterator, Generator

from ..core.context_manager import ObservabilityContext
from ..core.token_tracker import TokenTracker
from ..utils.helpers import count_tokens


class LiteLLMPatcher:
    """
    Patcher for LiteLLM's API that adds tracing, token tracking, and policy enforcement.

    Features:
    - Tracing with OpenTelemetry
    - Token tracking for prompt and response
    - Policy enforcement
    - Streaming support
    """

    def __init__(self,
                 token_tracker: TokenTracker = None,
                 log_file: Any = None,
                 context: ObservabilityContext = None):
        self._original_functions = {}
        self._token_tracker = token_tracker or TokenTracker()
        self._context = context or ObservabilityContext.get_current()

    def patch(self) -> None:
        """
        Patch LiteLLM's SDK functions.
        """
        try:
            import litellm

            pass

        except ImportError:
            raise ImportError("LiteLLM SDK is not installed.")

    def unpatch(self) -> None:
        """
        Restore original LiteLLM SDK functions.
        """
        try:
            import litellm

            pass

        except ImportError:
            raise ImportError("LiteLLM SDK is not installed.")



"""Utility functions for OpenTelemetry tracing"""

from typing import Dict, Any
from opentelemetry import trace
from opentelemetry.trace import Span
from observicia.core.context_manager import ObservabilityContext


def start_llm_span(name: str, attributes: Dict[str, Any]) -> Span:
    """Start a new trace span with LLM attributes."""
    tracer = trace.get_tracer(__name__)
    context = ObservabilityContext.get_current()

    # Combine base attributes with LLM-specific attributes
    span_attributes = {
        "service.name": context._service_name,
        "llm.provider": "openai",
        "llm.model": attributes.get('model', 'unknown'),
        "llm.request.type": name.split('.')[-1]
    }

    user_id = context.get_user_id()
    if user_id:
        span_attributes["user.id"] = user_id

    return tracer.start_span(name=name, attributes=span_attributes)


def record_token_usage(span: Span, response: Any) -> None:
    """Record token usage in span from response."""
    if hasattr(response, 'usage'):
        usage = response.usage
        usage_dict = {}

        # Handle different usage structures
        if hasattr(usage, 'prompt_tokens'):
            usage_dict['prompt.tokens'] = usage.prompt_tokens
        if hasattr(usage, 'completion_tokens'):
            usage_dict['completion.tokens'] = usage.completion_tokens
        elif hasattr(usage, 'total_tokens') and hasattr(
                usage, 'prompt_tokens'):
            # Some APIs only give total and prompt, derive completion
            usage_dict[
                'completion.tokens'] = usage.total_tokens - usage.prompt_tokens
        if hasattr(usage, 'total_tokens'):
            usage_dict['total.tokens'] = usage.total_tokens

        # Set all available usage metrics
        span.set_attributes(usage_dict)

        # print usage_dict
        print(f"Usage dict: {usage_dict}")



import asyncio
from functools import wraps
from typing import Any, AsyncIterator, Generator

from ..core.context_manager import ObservabilityContext
from ..core.token_tracker import TokenTracker
from ..utils.helpers import count_tokens


class AnthropicPatcher:
    """
    Patcher for Anthropic's API that adds tracing and token tracking.

    Features:
    - Tracing with OpenTelemetry
    - Token counting for prompt and response
    - Policy enforcement
    """

    def __init__(self,
                 token_tracker: TokenTracker = None,
                 log_file: Any = None,
                 context: ObservabilityContext = None):
        self._original_functions = {}
        self._token_tracker = token_tracker or TokenTracker()
        self._context = context or ObservabilityContext.get_current()

    def patch(self) -> None:
        """
        Patch Anthropic's SDK functions.
        """
        try:
            import anthropic

            pass

        except ImportError:
            raise ImportError("Anthropic SDK is not installed.")

    def unpatch(self) -> None:
        """
        Restore original Anthropic SDK functions.
        """
        try:
            import anthropic

            pass
        except ImportError:
            raise ImportError("Anthropic SDK is not installed.")



"""
Core components of the Observicia SDK.
"""

from .context_manager import ObservabilityContext
from .policy_engine import PolicyEngine, PolicyResult
from .tracing_manager import TracingClient
from .token_tracker import TokenTracker
from .patch_manager import PatchManager

__all__ = [
    "ObservabilityContext", "PolicyEngine", "PolicyResult", "TracingClient",
    "TokenTracker", "PatchManager"
]



import asyncio
from functools import wraps
from typing import Dict, Optional, Any
from contextlib import contextmanager

from observicia.core.context_manager import ObservabilityContext
from observicia.core.token_tracker import TokenTracker
from observicia.patchers import DEFAULT_PATCHERS


class PatchManager:
    """
    Manages patching and unpatching of LLM provider SDKs for token tracking and tracing.
    """

    def __init__(self) -> None:
        self._context = ObservabilityContext.get_current()
        self._token_tracker = TokenTracker()
        self._active_patches: Dict[str, Any] = {}
        self._original_functions: Dict[str, Any] = {}
        self._log_file = getattr(self._context, '_log_file', None)

    def patch_provider(self, provider_name: str) -> None:
        """
        Patch a specific LLM provider's SDK.
        
        Args:
            provider_name: Name of the LLM provider to patch (e.g., "openai", "anthropic")
        """
        if provider_name in self._active_patches:
            return  # Already patched

        patcher_class = DEFAULT_PATCHERS.get(provider_name)
        if not patcher_class:
            raise ValueError(f"Unsupported provider: {provider_name}")

        try:
            print(f"Patching {provider_name} SDK...")
            # Initialize patcher for this provider
            patcher = patcher_class(token_tracker=self._token_tracker,
                                    log_file=self._log_file,
                                    context=self._context)

            # Apply patches and store original functions
            orig_funcs = patcher.patch()
            self._original_functions[provider_name] = orig_funcs
            self._active_patches[provider_name] = patcher
            print(f"Patched {provider_name} SDK")

        except ImportError as e:
            raise ImportError(
                f"Could not patch {provider_name}. SDK not installed: {str(e)}"
            )
        except Exception as e:
            self._rollback_patches(provider_name)
            raise RuntimeError(f"Error patching {provider_name}: {str(e)}")

    def unpatch_provider(self, provider_name: str) -> None:
        """
        Remove patches for a specific provider.
        
        Args:
            provider_name: Name of the LLM provider to unpatch
        """
        if provider_name not in self._active_patches:
            return  # Not patched

        patcher = self._active_patches[provider_name]
        orig_funcs = self._original_functions[provider_name]

        try:
            patcher.unpatch(orig_funcs)
        finally:
            del self._active_patches[provider_name]
            del self._original_functions[provider_name]

    def patch_all(self) -> None:
        """Patch all supported LLM providers that are installed."""
        for provider in DEFAULT_PATCHERS.keys():
            try:
                self.patch_provider(provider)
            except ImportError:
                # Skip providers that aren't installed
                continue

    def unpatch_all(self) -> None:
        """Remove all active patches."""
        for provider in list(self._active_patches.keys()):
            self.unpatch_provider(provider)

    def _rollback_patches(self, provider_name: str) -> None:
        """Rollback patches for a provider if patching fails."""
        if provider_name in self._original_functions:
            patcher = self._active_patches.get(provider_name)
            if patcher:
                try:
                    patcher.unpatch(self._original_functions[provider_name])
                except Exception:
                    pass  # Best effort rollback
            del self._original_functions[provider_name]
            if provider_name in self._active_patches:
                del self._active_patches[provider_name]

    def __enter__(self):
        """Enable use as a context manager for automatic cleanup."""
        self.patch_all()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Cleanup when exiting context."""
        self.unpatch_all()



"""Helpers for serializing API responses."""
from typing import Any, Dict
from openai.types.chat import ChatCompletion, ChatCompletionMessage
from openai.types.completion import Completion


def serialize_chat_completion(response: ChatCompletion) -> Dict[str, Any]:
    """Convert ChatCompletion to a JSON-serializable dict."""
    return {
        "id":
        response.id,
        "choices": [{
            "index": choice.index,
            "message": {
                "role": choice.message.role,
                "content": choice.message.content
            } if choice.message else {},
            "finish_reason": choice.finish_reason
        } for choice in response.choices],
        "model":
        response.model,
        "usage": {
            "prompt_tokens": response.usage.prompt_tokens,
            "completion_tokens": response.usage.completion_tokens,
            "total_tokens": response.usage.total_tokens
        } if response.usage else {}
    }


def serialize_completion(response: Completion) -> Dict[str, Any]:
    """Convert Completion to a JSON-serializable dict."""
    return {
        "id":
        response.id,
        "choices": [{
            "text": choice.text,
            "index": choice.index,
            "finish_reason": choice.finish_reason
        } for choice in response.choices],
        "model":
        response.model,
        "usage": {
            "prompt_tokens": response.usage.prompt_tokens,
            "completion_tokens": response.usage.completion_tokens,
            "total_tokens": response.usage.total_tokens
        } if response.usage else {}
    }


def serialize_llm_response(response: Any) -> Dict[str, Any]:
    """Serialize any LLM response to a JSON-serializable format."""
    if isinstance(response, ChatCompletion):
        return serialize_chat_completion(response)
    elif isinstance(response, Completion):
        return serialize_completion(response)
    elif isinstance(response, dict):
        return response
    else:
        # For unknown types, try to convert to dict if possible
        try:
            return dict(response)
        except (TypeError, ValueError):
            return {"content": str(response)}



"""Utility functions for policy enforcement"""
from typing import Any, Optional
from opentelemetry.trace import Span
from observicia.core.context_manager import ObservabilityContext
from .serialization_helpers import serialize_llm_response


def enforce_policies(context: Optional[ObservabilityContext],
                     span: Span,
                     response: Any,
                     prompt: Optional[str] = None,
                     completion: Optional[str] = None) -> None:
    """Synchronously enforce policies on response."""
    if not context or not context.policy_engine:
        return

    # Serialize the response before evaluation
    serialized_response = serialize_llm_response(response)

    # Use synchronous evaluation
    result = context.policy_engine.evaluate_sync(
        {
            "response": serialized_response,
            "trace_context": {
                "trace_id": span.get_span_context().trace_id,
                "span_id": span.get_span_context().span_id,
                "attributes": dict(span.attributes)
            }
        },
        prompt=prompt,
        completion=completion)

    span.set_attributes({
        "policy.passed": result.passed,
        "policy.violations": ";".join(result.violations)
    })

    if not result.passed:
        raise ValueError(f"Policy violations: {result.violations}")



import pytest
from unittest.mock import Mock, patch, AsyncMock
from openai.resources.chat.completions import AsyncCompletions as AsyncChatCompletions
from openai.resources.chat.completions import Completions as ChatCompletions
from observicia.patchers.openai import OpenAIPatcher
from observicia.core.token_tracker import TokenTracker
from observicia.core.context_manager import ObservabilityContext


@pytest.fixture(autouse=True)
def setup_observability_context():
    """Initialize ObservabilityContext before each test."""
    ObservabilityContext._instance = None
    ObservabilityContext.initialize(service_name="test-service",
                                    trace_console=True)
    yield
    ObservabilityContext._instance = None


@pytest.fixture
def token_tracker():
    return TokenTracker()


@pytest.fixture
def mock_context(setup_observability_context):
    context = ObservabilityContext.get_current()
    context._service_name = "test-service"
    context.policy_engine = None
    return context


@pytest.fixture
def patcher(token_tracker, mock_context):
    return OpenAIPatcher(token_tracker=token_tracker, context=mock_context)


class TestOpenAIPatcher:

    def test_initialization(self, patcher):
        """Test patcher initialization."""
        assert isinstance(patcher._token_tracker, TokenTracker)
        assert len(patcher._original_functions) == 0
        assert not patcher._patched

    def test_patch_chat_completions(self, patcher):
        """Test patching chat completions."""
        original_chat_create = ChatCompletions.create

        # Apply patches
        patcher.patch()

        # Verify functions are patched
        assert ChatCompletions.create != original_chat_create
        assert patcher._patched

        # Clean up
        patcher.unpatch()
        assert not patcher._patched

    def test_unpatch(self, patcher):
        """Test unpatching functionality."""
        # Store original references
        original_chat_create = ChatCompletions.create
        original_async_chat_create = AsyncChatCompletions.create

        # Apply patches
        orig_funcs = patcher.patch()
        assert patcher._patched

        # Verify functions were patched
        assert ChatCompletions.create is not original_chat_create
        assert AsyncChatCompletions.create is not original_async_chat_create

        # Unpatch
        patcher.unpatch(orig_funcs)

        # Verify unpatching restored original functionality
        assert not patcher._patched
        current_chat_create = ChatCompletions.create
        assert current_chat_create.__name__ == original_chat_create.__name__
        assert current_chat_create.__module__ == original_chat_create.__module__

    def test_context_manager(self, patcher):
        """Test using patcher as context manager."""
        original_chat_create = ChatCompletions.create

        with patcher:
            assert patcher._patched
            assert ChatCompletions.create is not original_chat_create

        assert not patcher._patched
        current_chat_create = ChatCompletions.create
        assert current_chat_create.__name__ == original_chat_create.__name__
        assert current_chat_create.__module__ == original_chat_create.__module__

    def test_double_patch(self, patcher):
        """Test that patching twice doesn't cause issues."""
        original_chat_create = ChatCompletions.create

        # First patch
        patcher.patch()
        patched_func = ChatCompletions.create

        # Second patch
        patcher.patch()
        assert ChatCompletions.create is patched_func

        # Cleanup
        patcher.unpatch()
        current_chat_create = ChatCompletions.create
        assert current_chat_create.__name__ == original_chat_create.__name__
        assert current_chat_create.__module__ == original_chat_create.__module__


if __name__ == '__main__':
    pytest.main([__file__])



import threading
from collections import defaultdict
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import Dict, List, Optional


@dataclass
class TokenUsage:
    """Tracks token usage for a specific session or request."""
    prompt_tokens: int = 0
    completion_tokens: int = 0
    total_tokens: int = 0
    timestamp: datetime = field(default_factory=datetime.utcnow)


class TokenTracker:
    """token tracker for multiple LLM providers."""

    def __init__(self, retention_period: timedelta = timedelta(days=7)):
        self._lock = threading.Lock()
        self._retention_period = retention_period
        self._usage_by_provider: Dict[str,
                                      List[TokenUsage]] = defaultdict(list)
        self._stream_usage: Dict[str, TokenUsage] = {}

    def update(self, provider: str, prompt_tokens: int,
               completion_tokens: int) -> None:
        """Update token usage for a provider."""
        usage = TokenUsage(prompt_tokens=prompt_tokens,
                           completion_tokens=completion_tokens,
                           total_tokens=prompt_tokens + completion_tokens)

        with self._lock:
            self._usage_by_provider[provider].append(usage)
            self._cleanup_old_data(provider)

    def get_totals(self, provider: str) -> TokenUsage:
        """Retrieve total token usage for a provider."""
        with self._lock:
            total_usage = TokenUsage()
            for usage in self._usage_by_provider.get(provider, []):
                total_usage.prompt_tokens += usage.prompt_tokens
                total_usage.completion_tokens += usage.completion_tokens
                total_usage.total_tokens += usage.total_tokens
            return total_usage

    @contextmanager
    def stream_context(self, provider: str, session_id: str):
        """Context manager for streaming token tracking."""
        try:
            with self._lock:
                self._stream_usage[session_id] = TokenUsage()
            yield self._stream_usage[session_id]
        finally:
            with self._lock:
                if session_id in self._stream_usage:
                    final_usage = self._stream_usage[session_id]
                    usage = TokenUsage(
                        prompt_tokens=final_usage.prompt_tokens,
                        completion_tokens=final_usage.completion_tokens,
                        total_tokens=final_usage.prompt_tokens +
                        final_usage.completion_tokens)
                    self._usage_by_provider[provider].append(usage)
                    self._cleanup_old_data(provider)
                    del self._stream_usage[session_id]

    def get_usage(self,
                  provider: str,
                  window: Optional[timedelta] = None) -> Dict[str, int]:
        """Get token usage statistics for a provider."""
        with self._lock:
            usages = self._usage_by_provider[provider]
            if window:
                cutoff = datetime.utcnow() - window
                usages = [u for u in usages if u.timestamp >= cutoff]

            return {
                "prompt_tokens": sum(u.prompt_tokens for u in usages),
                "completion_tokens": sum(u.completion_tokens for u in usages),
                "total_tokens": sum(u.total_tokens for u in usages)
            }

    def get_usage_all_providers(
            self,
            window: Optional[timedelta] = None) -> Dict[str, Dict[str, int]]:
        """Get token usage statistics for all providers."""
        with self._lock:
            return {
                provider: self.get_usage(provider, window)
                for provider in self._usage_by_provider.keys()
            }

    def _cleanup_old_data(self, provider: str) -> None:
        """Remove data older than the retention period."""
        if provider not in self._usage_by_provider:
            return

        cutoff = datetime.utcnow() - self._retention_period
        self._usage_by_provider[provider] = [
            usage for usage in self._usage_by_provider[provider]
            if usage.timestamp >= cutoff
        ]



import pytest
import threading
from datetime import datetime, timedelta
from observicia.core.token_tracker import TokenTracker, TokenUsage


@pytest.fixture
def token_tracker():
    return TokenTracker(retention_period=timedelta(days=1))


def test_token_usage_initialization():
    """Test TokenUsage initialization and defaults."""
    usage = TokenUsage()
    assert usage.prompt_tokens == 0
    assert usage.completion_tokens == 0
    assert usage.total_tokens == 0
    assert isinstance(usage.timestamp, datetime)


def test_token_tracker_update(token_tracker):
    """Test updating token usage for a provider."""
    token_tracker.update("openai", prompt_tokens=10, completion_tokens=20)

    totals = token_tracker.get_totals("openai")
    assert totals.prompt_tokens == 10
    assert totals.completion_tokens == 20
    assert totals.total_tokens == 30


def test_stream_context(token_tracker):
    """Test token tracking in streaming context."""
    with token_tracker.stream_context("openai", "test_session") as usage:
        usage.prompt_tokens += 5
        usage.completion_tokens += 10

    totals = token_tracker.get_totals("openai")
    assert totals.prompt_tokens == 5
    assert totals.completion_tokens == 10


def test_stream_context_error_handling(token_tracker):
    """Test stream context handling when errors occur."""
    session_id = "test_session"

    try:
        with token_tracker.stream_context("openai", session_id) as usage:
            usage.prompt_tokens += 5
            usage.completion_tokens += 10
            raise ValueError("Test error")
    except ValueError:
        pass

    with token_tracker._lock:
        assert session_id not in token_tracker._stream_usage

    totals = token_tracker.get_totals("openai")
    assert totals.prompt_tokens == 5
    assert totals.completion_tokens == 10


def test_multi_provider(token_tracker):
    """Test tracking tokens for multiple providers."""
    token_tracker.update("openai", 10, 20)
    token_tracker.update("anthropic", 15, 25)
    token_tracker.update("openai", 5, 10)

    openai_totals = token_tracker.get_totals("openai")
    assert openai_totals.prompt_tokens == 15
    assert openai_totals.completion_tokens == 30

    anthropic_totals = token_tracker.get_totals("anthropic")
    assert anthropic_totals.prompt_tokens == 15
    assert anthropic_totals.completion_tokens == 25


def test_concurrent_updates(token_tracker):
    """Test thread safety with concurrent updates."""

    def update_tokens():
        for _ in range(100):
            token_tracker.update("openai", 1, 2)

    threads = []
    for _ in range(10):
        thread = threading.Thread(target=update_tokens)
        thread.start()
        threads.append(thread)

    for thread in threads:
        thread.join()

    totals = token_tracker.get_totals("openai")
    assert totals.prompt_tokens == 1000  # 10 threads * 100 updates * 1 token
    assert totals.completion_tokens == 2000  # 10 threads * 100 updates * 2 tokens


def test_retention_cleanup(token_tracker):
    """Test cleanup of old usage data."""
    old_usage = TokenUsage(prompt_tokens=10, completion_tokens=20)
    old_usage.timestamp = datetime.utcnow() - timedelta(days=2)

    with token_tracker._lock:
        token_tracker._usage_by_provider["test"].append(old_usage)

    token_tracker.update("test", 5, 10)  # This should trigger cleanup

    totals = token_tracker.get_totals("test")
    assert totals.prompt_tokens == 5  # Only recent usage remains
    assert totals.completion_tokens == 10


if __name__ == '__main__':
    pytest.main([__file__])



import pytest
from unittest.mock import Mock, patch, AsyncMock
from datetime import datetime
from opentelemetry import trace
from opentelemetry.trace import SpanKind, Status, StatusCode
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import SimpleSpanProcessor
from opentelemetry.sdk.trace.export import ConsoleSpanExporter

from observicia.core.context_manager import (ObservabilityContext,
                                             ContextManager, TraceContext)
from observicia.core.policy_engine import PolicyEngine, PolicyResult


@pytest.fixture
def mock_policy_engine():
    engine = Mock(spec=PolicyEngine)
    engine.evaluate_with_context = AsyncMock(
        return_value=PolicyResult(passed=True, violations=[]))
    return engine


@pytest.fixture
def context_manager():
    return ContextManager(service_name="test-service", trace_console=True)


@pytest.fixture
def trace_context():
    return TraceContext(trace_id="test-trace-id",
                        parent_id="test-parent-id",
                        attributes={"test_key": "test_value"},
                        session_id="test-session")


class TestTraceContext:

    def test_initialization(self):
        context = TraceContext(trace_id="test-trace",
                               parent_id="parent-id",
                               attributes={"key": "value"})

        assert context.trace_id == "test-trace"
        assert context.parent_id == "parent-id"
        assert context.attributes == {"key": "value"}
        assert context.violation_count == 0
        assert len(context.active_policies) == 0
        assert len(context.policy_results) == 0

    def test_to_dict(self):
        context = TraceContext(trace_id="test-trace",
                               parent_id="parent-id",
                               attributes={"key": "value"},
                               session_id="test-session")

        result = context.to_dict()
        assert result["trace_id"] == "test-trace"
        assert result["parent_id"] == "parent-id"
        assert result["session_id"] == "test-session"
        assert result["key"] == "value"
        assert isinstance(result["active_policies"], list)

    def test_add_violation(self):
        context = TraceContext(trace_id="test-trace",
                               parent_id="parent-id",
                               attributes={})

        policy_result = PolicyResult(
            passed=False, violations=["test violation 1", "test violation 2"])

        context.add_violation(policy_result)
        assert context.violation_count == 2
        assert len(context.policy_results) == 1
        assert context.policy_results[0] == policy_result

    def test_policy_management(self):
        context = TraceContext(trace_id="test-trace",
                               parent_id="parent-id",
                               attributes={})

        context.add_policy("test_policy")
        assert "test_policy" in context.active_policies

        context.remove_policy("test_policy")
        assert "test_policy" not in context.active_policies

        # Test removing non-existent policy
        context.remove_policy("non_existent")
        assert len(context.active_policies) == 0


class TestContextManager:

    def test_initialization(self):
        # Get the current tracer provider before creating context manager
        original_provider = trace.get_tracer_provider()

        manager = ContextManager(service_name="test-service",
                                 trace_console=True)

        # Verify that a tracer was created
        assert manager._tracer is not None
        assert manager._service_name == "test-service"

        # Verify that the provider was changed
        current_provider = trace.get_tracer_provider()
        assert isinstance(current_provider, TracerProvider)
        assert current_provider != original_provider

    def test_get_non_existent_session(self, context_manager):
        session = context_manager.get_session("non-existent")
        assert session is None

    def test_create_and_get_session(self, context_manager):
        session_id = "test-session"
        initial_context = {"test_key": "test_value"}

        context = context_manager.create_session(
            session_id=session_id, initial_context=initial_context)

        assert context.session_id == session_id
        assert context.attributes == initial_context

        # Verify we can retrieve the same session
        retrieved = context_manager.get_session(session_id)
        assert retrieved == context
        assert retrieved.attributes == initial_context

    @pytest.mark.asyncio
    async def test_create_span(self, context_manager, trace_context):
        span_name = "test-span"
        attributes = {"test_attr": "test_value"}

        span = await context_manager.create_span(name=span_name,
                                                 context=trace_context,
                                                 kind=SpanKind.CLIENT,
                                                 attributes=attributes)

        assert span is not None
        assert span.name == span_name
        assert span.kind == SpanKind.CLIENT

        # Check that context attributes were added to span
        span_attrs = dict(span.attributes)
        assert span_attrs["service.name"] == "test-service"
        assert span_attrs["test_attr"] == "test_value"
        assert span_attrs["trace_id"] == trace_context.trace_id

    @pytest.mark.asyncio
    async def test_policy_evaluation(self, context_manager, trace_context,
                                     mock_policy_engine):
        context_manager.policy_engine = mock_policy_engine
        eval_context = {"test": "context"}

        result = await context_manager.evaluate_policies(
            context=trace_context,
            eval_context=eval_context,
            policies=["test_policy"])

        assert result.passed is True
        assert len(result.violations) == 0
        mock_policy_engine.evaluate_with_context.assert_called_once_with(
            eval_context=eval_context, policies=["test_policy"])

    @pytest.mark.asyncio
    async def test_policy_evaluation_with_violations(self, context_manager,
                                                     trace_context,
                                                     mock_policy_engine):
        violation = "test violation"
        mock_policy_engine.evaluate_with_context.return_value = PolicyResult(
            passed=False, violations=[violation])

        context_manager.policy_engine = mock_policy_engine

        result = await context_manager.evaluate_policies(
            context=trace_context, eval_context={}, policies=["test_policy"])

        assert result.passed is False
        assert violation in result.violations
        assert trace_context.violation_count == 1


class TestObservabilityContext:

    def test_singleton_behavior(self):
        ObservabilityContext._instance = None  # Reset singleton

        # First initialization
        ObservabilityContext.initialize(service_name="test-service")
        first_instance = ObservabilityContext.get_current()

        # Second initialization shouldn't create new instance
        ObservabilityContext.initialize(service_name="another-service")
        second_instance = ObservabilityContext.get_current()

        assert first_instance is second_instance
        assert first_instance._service_name == "test-service"

    def test_uninitialized_access(self):
        ObservabilityContext._instance = None  # Reset singleton

        with pytest.raises(RuntimeError) as exc_info:
            ObservabilityContext.get_current()
        assert "ObservabilityContext not initialized" in str(exc_info.value)

    def test_create_session_uninitialized(self):
        ObservabilityContext._instance = None  # Reset singleton

        with pytest.raises(RuntimeError) as exc_info:
            ObservabilityContext.create_session("test-session")
        assert "ObservabilityContext not initialized" in str(exc_info.value)

    def test_initialization_with_options(self):
        ObservabilityContext._instance = None  # Reset singleton

        ObservabilityContext.initialize(service_name="test-service",
                                        otel_endpoint="http://localhost:4317",
                                        opa_endpoint="http://localhost:8181",
                                        trace_console=True)

        instance = ObservabilityContext.get_current()
        assert instance._service_name == "test-service"
        assert instance.policy_engine is not None


if __name__ == '__main__':
    pytest.main([__file__])



import pytest
from unittest.mock import Mock, patch, MagicMock
from observicia.core.patch_manager import PatchManager
from observicia.core.context_manager import ObservabilityContext
from observicia.core.token_tracker import TokenTracker


class MockPatcher:
    """Mock patcher class for testing."""

    def __init__(self, token_tracker=None, context=None):
        self.token_tracker = token_tracker
        self.context = context
        self.patched = False
        self.unpatched = False

    def patch(self):
        self.patched = True
        return {"original_func": lambda: None}

    def unpatch(self, original_funcs=None):  # Made original_funcs optional
        self.unpatched = True


@pytest.fixture(autouse=True)
def mock_observability_context():
    """Fixture to mock and initialize ObservabilityContext for all tests."""
    mock_context = Mock(spec=ObservabilityContext)
    mock_instance = Mock()
    mock_context._instance = mock_instance

    with patch(
            'observicia.core.patch_manager.ObservabilityContext') as mock_ctx:
        mock_ctx.get_current.return_value = mock_instance
        yield mock_ctx


@pytest.fixture
def token_tracker():
    return Mock(spec=TokenTracker)


@pytest.fixture
def patch_manager(token_tracker):
    return PatchManager()


def test_init(patch_manager):
    """Test PatchManager initialization."""
    assert patch_manager._active_patches == {}
    assert patch_manager._original_functions == {}
    assert isinstance(patch_manager._token_tracker, TokenTracker)


@pytest.mark.parametrize("provider_name",
                         ["openai", "anthropic", "litellm", "watsonx"])
def test_patch_provider_success(patch_manager, provider_name):
    """Test successful provider patching."""
    mock_patcher = MockPatcher()

    with patch('observicia.core.patch_manager.DEFAULT_PATCHERS',
               {provider_name: lambda **kwargs: mock_patcher}):
        patch_manager.patch_provider(provider_name)

        assert provider_name in patch_manager._active_patches
        assert provider_name in patch_manager._original_functions
        assert mock_patcher.patched


def test_patch_provider_unsupported(mock_observability_context):
    """Test patching an unsupported provider."""
    patch_manager = PatchManager()
    with pytest.raises(ValueError, match="Unsupported provider"):
        patch_manager.patch_provider("unsupported_provider")


def test_patch_provider_import_error(mock_observability_context):
    """Test handling of import errors during patching."""
    patch_manager = PatchManager()

    def mock_patcher(*args, **kwargs):
        raise ImportError("SDK not installed")

    with patch('observicia.core.patch_manager.DEFAULT_PATCHERS',
               {"test": mock_patcher}):
        with pytest.raises(ImportError, match="SDK not installed"):
            patch_manager.patch_provider("test")


def test_patch_provider_runtime_error(mock_observability_context):
    """Test handling of runtime errors during patching."""
    patch_manager = PatchManager()

    def mock_patcher(*args, **kwargs):
        raise RuntimeError("Patching failed")

    with patch('observicia.core.patch_manager.DEFAULT_PATCHERS',
               {"test": mock_patcher}):
        with pytest.raises(RuntimeError, match="Error patching test"):
            patch_manager.patch_provider("test")


def test_unpatch_provider(patch_manager):
    """Test unpatching a provider."""
    mock_patcher = MockPatcher()
    provider_name = "test"

    # Setup patched state
    patch_manager._active_patches[provider_name] = mock_patcher
    patch_manager._original_functions[provider_name] = {"func": lambda: None}

    patch_manager.unpatch_provider(provider_name)

    assert provider_name not in patch_manager._active_patches
    assert provider_name not in patch_manager._original_functions
    assert mock_patcher.unpatched


def test_unpatch_provider_not_patched(patch_manager):
    """Test unpatching a provider that wasn't patched."""
    patch_manager.unpatch_provider("test")  # Should not raise any exceptions


def test_patch_all(patch_manager):
    """Test patching all providers."""
    mock_patcher = MockPatcher()
    providers = ["openai", "anthropic"]

    with patch(
            'observicia.core.patch_manager.DEFAULT_PATCHERS',
        {provider: lambda **kwargs: mock_patcher
         for provider in providers}):
        patch_manager.patch_all()

        for provider in providers:
            assert provider in patch_manager._active_patches
            assert provider in patch_manager._original_functions


def test_unpatch_all(patch_manager):
    """Test unpatching all providers."""
    mock_patcher = MockPatcher()
    providers = ["openai", "anthropic"]

    # Setup patched state
    for provider in providers:
        patch_manager._active_patches[provider] = mock_patcher
        patch_manager._original_functions[provider] = {"func": lambda: None}

    patch_manager.unpatch_all()

    assert len(patch_manager._active_patches) == 0
    assert len(patch_manager._original_functions) == 0
    assert mock_patcher.unpatched


def test_context_manager(mock_observability_context):
    """Test PatchManager as context manager."""
    mock_patcher = MockPatcher()
    providers = ["openai", "anthropic"]

    with patch(
            'observicia.core.patch_manager.DEFAULT_PATCHERS',
        {provider: lambda **kwargs: mock_patcher
         for provider in providers}):
        with PatchManager() as manager:
            assert all(provider in manager._active_patches
                       for provider in providers)

        # After context exit, should be unpatched
        assert len(manager._active_patches) == 0
        assert len(manager._original_functions) == 0


def test_rollback_patches(patch_manager):
    """Test rollback functionality when patching fails."""
    mock_patcher = MockPatcher()
    provider = "test"

    # Setup initial state
    patch_manager._original_functions[provider] = {"func": lambda: None}
    patch_manager._active_patches[provider] = mock_patcher

    patch_manager._rollback_patches(provider)

    assert provider not in patch_manager._active_patches
    assert provider not in patch_manager._original_functions
    assert mock_patcher.unpatched


@pytest.mark.asyncio
async def test_async_compatibility(mock_observability_context):
    """Test PatchManager compatibility with async code."""
    mock_patcher = MockPatcher()

    with patch('observicia.core.patch_manager.DEFAULT_PATCHERS',
               {"test": lambda **kwargs: mock_patcher}):

        async def async_function():
            with PatchManager() as manager:
                return manager

        manager = await async_function()
        assert isinstance(manager, PatchManager)


def test_double_patch(patch_manager):
    """Test that patching an already patched provider doesn't cause issues."""
    mock_patcher = MockPatcher()
    provider = "test"

    with patch('observicia.core.patch_manager.DEFAULT_PATCHERS',
               {provider: lambda **kwargs: mock_patcher}):
        # Patch twice
        patch_manager.patch_provider(provider)
        patch_manager.patch_provider(provider)

        # Should only be patched once
        assert len(patch_manager._active_patches) == 1
        assert len(patch_manager._original_functions) == 1


def test_unpatch_during_exception(patch_manager):
    """Test that providers are properly unpatched even if an exception occurs."""
    mock_patcher = MockPatcher()
    provider = "test"

    class TestException(Exception):
        pass

    with patch('observicia.core.patch_manager.DEFAULT_PATCHERS',
               {provider: lambda **kwargs: mock_patcher}):
        try:
            with PatchManager() as manager:
                raise TestException("Test exception")
        except TestException:
            pass

        # Should be unpatched despite exception
        assert len(manager._active_patches) == 0
        assert len(manager._original_functions) == 0



"""
Test suite for the Observicia SDK.
"""

import os
import sys

# Add SDK root to Python path for test imports
sys.path.insert(0,
                os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))



"""
Utility functions and helpers for the Observicia SDK.
Provides common functionality used across the SDK components.
"""

import json
import re
import tiktoken
from typing import Any, Dict, Optional, Union
from datetime import datetime
from contextvars import ContextVar
import numpy as np

from opentelemetry.trace import get_current_span as otel_get_current_span
from opentelemetry.trace import Span, SpanContext
from opentelemetry.baggage import get_all as get_baggage

# Type alias for trace attributes
TraceAttributes = Dict[str, Union[str, int, float, bool]]


def get_current_span() -> Optional[Span]:
    """
    Get the current active span from the trace context.
    
    Returns:
        Optional[Span]: The current span or None if no span is active
    """
    span = otel_get_current_span()
    if span and span.is_recording():
        return span
    return None


def get_current_context() -> Dict[str, Any]:
    """
    Get the current context including trace and baggage information.
    
    Returns:
        Dict[str, Any]: Dictionary containing current context information
    """
    span = get_current_span()
    context = {
        "trace_id": None,
        "span_id": None,
        "baggage": get_baggage(),
        "timestamp": datetime.utcnow().isoformat()
    }

    if span:
        span_context: SpanContext = span.get_span_context()
        context.update({
            "trace_id": format_trace_id(span_context.trace_id),
            "span_id": format_span_id(span_context.span_id)
        })

    return context


def count_tokens(text: str, model: str = "gpt-3.5-turbo") -> int:
    """
    Count tokens in text using the appropriate tokenizer for the model.
    
    Args:
        text (str): Text to count tokens for
        model (str): Model name to determine tokenizer
        
    Returns:
        int: Number of tokens in the text
    """
    try:
        encoding = tiktoken.encoding_for_model(model)
        return len(encoding.encode(text))
    except KeyError:
        # Fallback to cl100k_base for unknown models
        encoding = tiktoken.get_encoding("cl100k_base")
        return len(encoding.encode(text))
    except Exception as e:
        # Log error and return approximate token count
        print(f"Error counting tokens: {e}")
        return len(text.split())  # Rough approximation


def format_trace_attributes(attributes: Dict[str, Any]) -> TraceAttributes:
    """
    Format attributes to be compatible with OpenTelemetry trace attributes.
    
    Args:
        attributes (Dict[str, Any]): Raw attributes to format
        
    Returns:
        TraceAttributes: Formatted attributes safe for tracing
    """
    formatted = {}

    for key, value in attributes.items():
        # Ensure key is a string
        key = str(key)

        # Handle different value types
        if isinstance(value, (str, int, float, bool)):
            formatted[key] = value
        elif isinstance(value, (datetime, np.ndarray)):
            formatted[key] = str(value)
        elif isinstance(value, dict):
            formatted[key] = json.dumps(value, default=str)
        elif value is None:
            formatted[key] = "null"
        else:
            formatted[key] = str(value)

    return formatted


def safe_json_serialize(obj: Any) -> str:
    """
    Safely serialize any object to JSON string.
    
    Args:
        obj (Any): Object to serialize
        
    Returns:
        str: JSON string representation of the object
    """

    def default(o: Any) -> str:
        if isinstance(o, (datetime, np.ndarray)):
            return str(o)
        return f"<non-serializable: {type(o).__name__}>"

    try:
        return json.dumps(obj, default=default)
    except Exception as e:
        return f"<serialization-error: {str(e)}>"


def format_trace_id(trace_id: int) -> str:
    """
    Format trace ID as hexadecimal string.
    
    Args:
        trace_id (int): Trace ID as integer
        
    Returns:
        str: Formatted trace ID
    """
    return f"{trace_id:032x}"


def format_span_id(span_id: int) -> str:
    """
    Format span ID as hexadecimal string.
    
    Args:
        span_id (int): Span ID as integer
        
    Returns:
        str: Formatted span ID
    """
    return f"{span_id:016x}"


def sanitize_key(key: str) -> str:
    """
    Sanitize attribute key for use in traces.
    
    Args:
        key (str): Raw attribute key
        
    Returns:
        str: Sanitized key
    """
    # Replace invalid characters with underscore
    sanitized = re.sub(r'[^a-zA-Z0-9._-]', '_', key)
    # Ensure key starts with letter or underscore
    if not sanitized[0].isalpha() and sanitized[0] != '_':
        sanitized = f"_{sanitized}"
    return sanitized


def truncate_string(s: str, max_length: int = 1000) -> str:
    """
    Truncate string to maximum length.
    
    Args:
        s (str): String to truncate
        max_length (int): Maximum length
        
    Returns:
        str: Truncated string
    """
    if len(s) <= max_length:
        return s
    return f"{s[:max_length-3]}..."


class MetricsHelper:
    """Helper class for handling metrics calculations and formatting."""

    @staticmethod
    def calculate_rate(count: int, duration_seconds: float) -> float:
        """Calculate rate per second."""
        if duration_seconds <= 0:
            return 0.0
        return count / duration_seconds

    @staticmethod
    def format_duration(duration_seconds: float) -> str:
        """Format duration in human readable format."""
        if duration_seconds < 1:
            return f"{duration_seconds*1000:.2f}ms"
        elif duration_seconds < 60:
            return f"{duration_seconds:.2f}s"
        else:
            minutes = int(duration_seconds / 60)
            seconds = duration_seconds % 60
            return f"{minutes}m {seconds:.2f}s"

    @staticmethod
    def calculate_percentile(values: list, percentile: float) -> float:
        """Calculate percentile of values."""
        if not values:
            return 0.0
        return float(np.percentile(values, percentile))


def get_memory_usage() -> Dict[str, float]:
    """
    Get current memory usage statistics.
    
    Returns:
        Dict[str, float]: Memory usage metrics in MB
    """
    import psutil
    process = psutil.Process()
    memory_info = process.memory_info()

    return {
        "rss_mb": memory_info.rss / (1024 * 1024),
        "vms_mb": memory_info.vms / (1024 * 1024)
    }


def exponential_backoff(attempt: int, base_delay: float = 1.0) -> float:
    """
    Calculate exponential backoff delay.
    
    Args:
        attempt (int): Attempt number (0-based)
        base_delay (float): Base delay in seconds
        
    Returns:
        float: Delay in seconds
    """
    return min(base_delay * (2**attempt), 30.0)  # Cap at 30 seconds


# Global context for thread-local storage
_thread_local = ContextVar("observicia_context", default={})


def get_thread_local() -> Dict[str, Any]:
    """Get thread-local storage dictionary."""
    return _thread_local.get()


def set_thread_local(key: str, value: Any) -> None:
    """Set value in thread-local storage."""
    context = _thread_local.get()
    context[key] = value
    _thread_local.set(context)


def clear_thread_local() -> None:
    """Clear thread-local storage."""
    _thread_local.set({})



import functools
from typing import Any, Dict

from ..core.context_manager import ObservabilityContext
from ..core.token_tracker import TokenTracker
from ..utils.helpers import count_tokens


class WatsonxPatcher:
    """
    Patcher for IBM's ibm-watsonx-ai SDK that adds tracing, token tracking, and policy enforcement.
    """

    def __init__(self,
                 token_tracker: TokenTracker = None,
                 log_file: Any = None,
                 context: ObservabilityContext = None):
        self._original_functions = {}
        self._token_tracker = token_tracker or TokenTracker()
        self._context = context or ObservabilityContext.get_current()

    def patch(self) -> None:
        """
        Patch the ibm-watsonx-ai SDK functions.
        """
        try:
            import ibm_watsonx_ai

            # Patch the inference function
            pass
        except ImportError:
            raise ImportError("ibm-watsonx-ai SDK is not installed.")

    def unpatch(self) -> None:
        """
        Restore original ibm-watsonx-ai SDK functions.
        """
        try:
            import ibm_watsonx_ai

            pass

        except ImportError:
            raise ImportError("ibm-watsonx-ai SDK is not installed.")



"""
Observicia - Policy-Aware Tracing SDK for LLM Applications
"""

from observicia.core.context_manager import ObservabilityContext
from observicia.core.policy_engine import PolicyEngine, PolicyResult
from observicia.core.tracing_manager import TracingClient
from observicia.core.token_tracker import TokenTracker
from observicia.core.patch_manager import PatchManager
from observicia.core.policy_engine import Policy
from typing import List, Optional
import os
import yaml

__version__ = "0.1.4"


def init() -> None:
    """Initialize the Observicia SDK.
    """
    # Get the config file path from an environment variable
    config_file = os.environ.get("OBSERVICIA_CONFIG_FILE",
                                 "observicia_config.yaml")

    # Load configurations
    try:
        with open(config_file, "r") as file:
            config = yaml.safe_load(file)
            # Extract configurations
            service_name = config.get("service_name", "default-service")
            otel_endpoint = config.get("otel_endpoint", None)
            opa_endpoint = config.get("opa_endpoint", None)
            policies = config.get("policies", [])
            log_file = config.get("log_file", None)
            trace_console = config.get("trace_console", False)

            policy_objects = [Policy(**policy)
                              for policy in policies] if policies else None

            ObservabilityContext.initialize(service_name=service_name,
                                            otel_endpoint=otel_endpoint,
                                            opa_endpoint=opa_endpoint,
                                            trace_console=trace_console,
                                            log_file=log_file,
                                            policies=policy_objects)

            # Auto-detect and patch installed providers
            patch_manager = PatchManager()
            patch_manager.patch_all()

    except FileNotFoundError:
        print(f"Configuration file {config_file} not found. Ignoring.")
    except yaml.YAMLError as e:
        print(f"Error parsing the YAML configuration file: {e}. Ignoring.")


# Expose main decorators
from observicia.core.tracing_manager import trace, trace_rag, trace_stream

# Expose utilities
from observicia.utils.helpers import get_current_span, get_current_context

__all__ = [
    "init", "trace", "trace_rag", "trace_stream", "get_current_span",
    "get_current_context", "PolicyResult", "__version__"
]



import pytest
from datetime import datetime
import json
from opentelemetry.trace import SpanKind
from observicia.utils.helpers import (
    count_tokens,
    format_trace_attributes,
    safe_json_serialize,
    format_trace_id,
    format_span_id,
    sanitize_key,
    truncate_string,
    MetricsHelper,
)
from observicia.utils.token_helpers import count_prompt_tokens, count_text_tokens
from observicia.utils.serialization_helpers import (
    serialize_chat_completion,
    serialize_completion,
    serialize_llm_response,
)


def test_count_tokens():
    """Test token counting with different models."""
    text = "Hello, world!"

    # Test with standard model
    count = count_tokens(text, "gpt-3.5-turbo")
    assert count > 0

    # Test with unknown model (should use fallback)
    count_unknown = count_tokens(text, "unknown-model")
    assert count_unknown > 0


def test_format_trace_attributes():
    """Test attribute formatting for traces."""
    attributes = {
        "string": "value",
        "int": 42,
        "float": 3.14,
        "bool": True,
        "datetime": datetime.now(),
        "dict": {
            "key": "value"
        },
        "none": None
    }

    formatted = format_trace_attributes(attributes)

    assert isinstance(formatted["string"], str)
    assert isinstance(formatted["int"], int)
    assert isinstance(formatted["float"], float)
    assert isinstance(formatted["bool"], bool)
    assert isinstance(formatted["datetime"], str)
    assert isinstance(formatted["dict"], str)
    assert formatted["none"] == "null"


def test_safe_json_serialize():
    """Test JSON serialization with various types."""
    data = {
        "datetime": datetime.now(),
        "regular": "string",
        "nested": {
            "key": "value"
        }
    }

    serialized = safe_json_serialize(data)
    assert isinstance(serialized, str)

    # Test with non-serializable object
    class NonSerializable:
        pass

    data_with_nonserializable = {"obj": NonSerializable()}
    result = safe_json_serialize(data_with_nonserializable)
    assert "non-serializable" in result.lower()


def test_trace_id_formatting():
    """Test trace ID formatting."""
    trace_id = 123456789
    formatted = format_trace_id(trace_id)
    assert len(formatted) == 32
    assert all(c in '0123456789abcdef' for c in formatted)


def test_span_id_formatting():
    """Test span ID formatting."""
    span_id = 123456789
    formatted = format_span_id(span_id)
    assert len(formatted) == 16
    assert all(c in '0123456789abcdef' for c in formatted)


def test_key_sanitization():
    """Test attribute key sanitization."""
    assert sanitize_key("normal_key") == "normal_key"
    assert sanitize_key("123key") == "_123key"
    assert sanitize_key("special!@#key") == "special___key"


def test_string_truncation():
    """Test string truncation functionality."""
    long_string = "a" * 2000
    truncated = truncate_string(long_string, max_length=1000)
    assert len(truncated) == 1000
    assert truncated.endswith("...")

    # Test string shorter than max length
    short_string = "Hello, world!"
    assert truncate_string(short_string, max_length=1000) == short_string


class TestMetricsHelper:
    """Test the MetricsHelper class methods."""

    def test_calculate_rate(self):
        """Test rate calculation."""
        assert MetricsHelper.calculate_rate(100, 10) == 10.0
        assert MetricsHelper.calculate_rate(100, 0) == 0.0

    def test_format_duration(self):
        """Test duration formatting."""
        assert MetricsHelper.format_duration(0.5) == "500.00ms"
        assert MetricsHelper.format_duration(5) == "5.00s"
        assert MetricsHelper.format_duration(65) == "1m 5.00s"

    def test_calculate_percentile(self):
        """Test percentile calculation."""
        values = [1, 2, 3, 4, 5]
        assert MetricsHelper.calculate_percentile(values, 50) == 3.0
        assert MetricsHelper.calculate_percentile([], 50) == 0.0


def test_token_counting():
    """Test token counting functions."""
    # Test prompt token counting
    messages = [{
        "role": "user",
        "content": "Hello"
    }, {
        "role": "assistant",
        "content": "Hi there"
    }]
    prompt_tokens = count_prompt_tokens(messages, "gpt-3.5-turbo")
    assert prompt_tokens > 0

    # Test text token counting
    text = "Hello, world!"
    text_tokens = count_text_tokens(text, "gpt-3.5-turbo")
    assert text_tokens > 0


def test_response_serialization():
    """Test response serialization functions."""

    # Mock response objects
    class MockChoice:

        def __init__(self, text=None, message=None):
            self.text = text
            self.message = message
            self.index = 0
            self.finish_reason = "stop"

    class MockUsage:

        def __init__(self):
            self.prompt_tokens = 10
            self.completion_tokens = 20
            self.total_tokens = 30

    class MockMessage:

        def __init__(self):
            self.role = "assistant"
            self.content = "Hello"

    class MockChatCompletion:

        def __init__(self):
            self.id = "test_id"
            message = MockMessage()
            self.choices = [MockChoice(message=message)]
            self.model = "gpt-3.5-turbo"
            self.usage = MockUsage()

    class MockCompletion:

        def __init__(self):
            self.id = "test_id"
            self.choices = [MockChoice(text="Hello")]
            self.model = "text-davinci-003"
            self.usage = MockUsage()

    # Test chat completion serialization
    chat_completion = MockChatCompletion()
    serialized_chat = serialize_chat_completion(chat_completion)
    assert serialized_chat["id"] == "test_id"
    assert serialized_chat["choices"][0]["message"]["content"] == "Hello"

    # Test completion serialization
    completion = MockCompletion()
    serialized = serialize_completion(completion)
    assert serialized["id"] == "test_id"
    assert serialized["choices"][0]["text"] == "Hello"

    # Test generic response serialization
    assert isinstance(serialize_llm_response(completion), dict)
    assert isinstance(serialize_llm_response(chat_completion), dict)
    assert isinstance(serialize_llm_response({"key": "value"}), dict)



"""
logging utilities for Observicia SDK.
"""

import logging
import sys
from datetime import datetime
from typing import Any, Dict, Optional, Union
from opentelemetry.trace import get_current_span, SpanContext
from opentelemetry._logs import set_logger_provider
from opentelemetry.sdk._logs import LoggerProvider
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor, ConsoleLogExporter
from opentelemetry.sdk.trace.export import SpanExporter
from opentelemetry.sdk.trace import ReadableSpan
from typing import Sequence

import json


class ConsoleFormatter(logging.Formatter):
    """
    Custom formatter for console and file output that includes trace context and timestamps.
    """

    def format(self, record: logging.LogRecord) -> str:
        timestamp = datetime.fromtimestamp(record.created).isoformat()

        # Get trace context if available
        trace_context = getattr(record, 'trace_context', {})
        trace_id = trace_context.get('trace_id', '')
        span_id = trace_context.get('span_id', '')

        # Format the basic message
        log_entry = {
            'timestamp': timestamp,
            'level': record.levelname,
            'name': record.name,
            'message': record.getMessage(),
        }

        # Add trace context if available
        if trace_id:
            log_entry['trace_id'] = trace_id
        if span_id:
            log_entry['span_id'] = span_id

        # Add extra fields if available
        extra = getattr(record, 'extra', None)
        if extra:
            log_entry['extra'] = extra

        # Add exception info if available
        if record.exc_info:
            log_entry['exception'] = self.formatException(record.exc_info)

        return json.dumps(log_entry, indent=2)


class FileSpanExporter(SpanExporter):
    """
    SpanExporter that writes spans to a file in JSON format.
    """

    def __init__(self, file_path: str):
        self.file_path = file_path
        # Configure file logger for spans
        self.logger = logging.getLogger("SpanExporter")
        self.logger.setLevel(logging.INFO)

        # Create file handler
        file_handler = logging.FileHandler(file_path)
        file_handler.setFormatter(
            logging.Formatter('%(message)s')  # Raw message for JSON
        )
        self.logger.addHandler(file_handler)

        # Remove any existing handlers to avoid duplicate output
        for handler in self.logger.handlers[:-1]:
            self.logger.removeHandler(handler)

    def _extract_span_data(self, span: ReadableSpan) -> Dict[str, Any]:
        """Extract relevant data from a span for serialization."""
        ctx = span.get_span_context()

        # Convert attributes to serializable format
        attributes = {}
        for key, value in span.attributes.items():
            if isinstance(value, (str, int, float, bool)):
                attributes[key] = value
            else:
                attributes[key] = str(value)

        # Extract events
        events = []
        for event in span.events:
            event_data = {
                "name": event.name,
                "timestamp": event.timestamp,
                "attributes": {
                    k: str(v) if not isinstance(v,
                                                (str, int, float, bool)) else v
                    for k, v in event.attributes.items()
                }
            }
            events.append(event_data)

        # Build span data structure
        span_data = {
            "type": "span",
            "timestamp": datetime.utcnow().isoformat(),
            "name": span.name,
            "trace_id": format(ctx.trace_id, "032x"),
            "span_id": format(ctx.span_id, "016x"),
            "parent_id":
            format(span.parent.span_id, "016x") if span.parent else None,
            "start_time": span.start_time,
            "end_time": span.end_time,
            "attributes": attributes,
            "status": {
                "status_code": span.status.status_code.name,
                "description": span.status.description
            },
            "events": events
        }

        return span_data

    def export(self, spans: Sequence[ReadableSpan]) -> None:
        """Export the spans to file."""
        try:
            for span in spans:
                span_data = self._extract_span_data(span)
                self.logger.info(json.dumps(span_data))
            return None
        except Exception as e:
            print(f"Error exporting spans to file: {e}")
            return None

    def force_flush(self, timeout_millis: float = 30000) -> bool:
        """Force flush the exporter."""
        return True

    def shutdown(self) -> None:
        """Shutdown the exporter."""
        for handler in self.logger.handlers:
            handler.close()


class ObserviciaLogger:
    """
    Logger class that supports console, file, and OpenTelemetry logging.
    """

    def __init__(self,
                 service_name: str,
                 log_level: int = logging.INFO,
                 console_output: bool = False,
                 file_output: Optional[str] = None,
                 otlp_endpoint: Optional[str] = None):
        """
        Initialize the logger with multiple output options.
        
        Args:
            service_name: Name of the service
            log_level: Logging level (default: INFO)
            console_output: Whether to output to console
            file_output: File path for log output (optional)
            otlp_endpoint: OpenTelemetry endpoint (optional)
        """
        self.logger = logging.getLogger(service_name)
        self.logger.setLevel(log_level)

        # Clear any existing handlers
        self.logger.handlers = []

        # Create formatter
        formatter = ConsoleFormatter()

        if console_output:
            # Console handler
            console_handler = logging.StreamHandler(sys.stdout)
            console_handler.setFormatter(formatter)
            self.logger.addHandler(console_handler)

        if file_output:
            # File handler
            file_handler = logging.FileHandler(file_output)
            file_handler.setFormatter(formatter)
            self.logger.addHandler(file_handler)

        if otlp_endpoint:
            # OpenTelemetry setup
            logger_provider = LoggerProvider()
            set_logger_provider(logger_provider)

            # Console exporter for OpenTelemetry
            console_exporter = ConsoleLogExporter()
            logger_provider.add_log_record_processor(
                BatchLogRecordProcessor(console_exporter))

    def _get_trace_context(self) -> Dict[str, str]:
        """Get current trace context if available."""
        span = get_current_span()
        if span:
            ctx: SpanContext = span.get_span_context()
            return {
                "trace_id": format(ctx.trace_id, "032x"),
                "span_id": format(ctx.span_id, "016x"),
            }
        return {}

    def _log(self,
             level: int,
             message: str,
             extra: Optional[Dict[str, Any]] = None,
             exc_info: Any = None) -> None:
        """
        Log a message with trace context and additional attributes.
        """
        trace_context = self._get_trace_context()

        # Create record extras
        extra_dict = {}
        if trace_context:
            extra_dict['trace_context'] = trace_context
        if extra:
            extra_dict['extra'] = extra

        self.logger.log(level, message, extra=extra_dict, exc_info=exc_info)

    def debug(self,
              message: str,
              extra: Optional[Dict[str, Any]] = None) -> None:
        self._log(logging.DEBUG, message, extra)

    def info(self,
             message: str,
             extra: Optional[Dict[str, Any]] = None) -> None:
        self._log(logging.INFO, message, extra)

    def warning(self,
                message: str,
                extra: Optional[Dict[str, Any]] = None) -> None:
        self._log(logging.WARNING, message, extra)

    def error(self,
              message: str,
              extra: Optional[Dict[str, Any]] = None,
              exc_info: Any = None) -> None:
        self._log(logging.ERROR, message, extra, exc_info)

    def critical(self,
                 message: str,
                 extra: Optional[Dict[str, Any]] = None,
                 exc_info: Any = None) -> None:
        self._log(logging.CRITICAL, message, extra, exc_info)



import functools
import os
from typing import Any, Dict, Optional, Union, AsyncGenerator

from openai import AsyncOpenAI, OpenAI
from openai.resources.chat.completions import AsyncCompletions as AsyncChatCompletions
from openai.resources.chat.completions import Completions as ChatCompletions
from openai.resources.completions import Completions, AsyncCompletions
from openai.resources.embeddings import AsyncEmbeddings, Embeddings
from openai.resources.files import AsyncFiles, Files
from openai.resources.images import AsyncImages, Images
from openai.types.chat import ChatCompletion, ChatCompletionChunk

from ..core.token_tracker import TokenTracker
from ..core.context_manager import ObservabilityContext
from ..utils.tracing_helpers import start_llm_span, record_token_usage
from ..utils.token_helpers import count_prompt_tokens, count_text_tokens, update_token_usage
from ..utils.policy_helpers import enforce_policies
from ..utils.stream_helpers import handle_async_stream, handle_stream
from ..utils.logging import ObserviciaLogger


class OpenAIPatcher:
    """Enhanced patcher for OpenAI's v1.0+ API with observability features."""

    def __init__(self,
                 token_tracker: Optional[TokenTracker] = None,
                 log_file: Optional[str] = None,
                 context: Optional[ObservabilityContext] = None):
        self._original_functions: Dict[str, Any] = {}
        self._token_tracker = token_tracker or TokenTracker()
        self._context = context or ObservabilityContext.get_current()
        self._patched = False
        self.logger = ObserviciaLogger("OpenAIPatcher",
                                       console_output=False,
                                       file_output=log_file,
                                       otlp_endpoint=None)

    def patch(self) -> Dict[str, Any]:
        """Apply patches to OpenAI SDK functions."""
        if self._patched:
            return self._original_functions

        try:
            original_functions = {
                'chat_completions.create': ChatCompletions.create,
                'async_chat_completions.create': AsyncChatCompletions.create
            }

            # Patch completions
            if hasattr(AsyncCompletions, "create"):
                original_functions[
                    "async_completions.create"] = AsyncCompletions.create
                AsyncCompletions.create = self._wrap_async_completion(
                    AsyncCompletions.create)

            if hasattr(Completions, "create"):
                original_functions["completions.create"] = Completions.create
                Completions.create = self._wrap_completion(Completions.create)

            # Patch chat completions
            if hasattr(AsyncChatCompletions, "create"):
                original_functions[
                    "async_chat_completions.create"] = AsyncChatCompletions.create
                AsyncChatCompletions.create = self._wrap_async_chat_completion(
                    AsyncChatCompletions.create)

            if hasattr(ChatCompletions, "create"):
                original_functions[
                    "chat_completions.create"] = ChatCompletions.create
                ChatCompletions.create = self._wrap_chat_completion(
                    ChatCompletions.create)

            # Patch embeddings
            if hasattr(AsyncEmbeddings, "create"):
                original_functions[
                    "async_embeddings.create"] = AsyncEmbeddings.create
                AsyncEmbeddings.create = self._wrap_async_embedding(
                    AsyncEmbeddings.create)

            if hasattr(Embeddings, "create"):
                original_functions["embeddings.create"] = Embeddings.create
                Embeddings.create = self._wrap_embedding(Embeddings.create)

            # Patch files
            if hasattr(AsyncFiles, "create"):
                original_functions["async_files.create"] = AsyncFiles.create
                AsyncFiles.create = self._wrap_async_file_upload(
                    AsyncFiles.create)

            if hasattr(Files, "create"):
                original_functions["files.create"] = Files.create
                Files.create = self._wrap_file_upload(Files.create)

            # Patch images
            if hasattr(AsyncImages, "generate"):
                original_functions[
                    "async_images.generate"] = AsyncImages.generate
                AsyncImages.generate = self._wrap_async_image_generation(
                    AsyncImages.generate)

            if hasattr(Images, "generate"):
                original_functions["images.generate"] = Images.generate
                Images.generate = self._wrap_image_generation(Images.generate)

            self._original_functions = original_functions
            self._patched = True
            return original_functions

        except ImportError as e:
            raise ImportError(f"OpenAI SDK not installed: {str(e)}")
        except Exception as e:
            self._rollback_patches()
            raise RuntimeError(f"Error patching OpenAI SDK: {str(e)}")

    def unpatch(self,
                original_functions: Optional[Dict[str, Any]] = None) -> None:
        """Restore original OpenAI SDK functions."""
        if not self._patched:
            return

        try:
            original_functions = original_functions or self._original_functions
            for func_path, original_func in original_functions.items():
                module_path, func_name = func_path.split('.')
                if hasattr(globals().get(module_path), func_name):
                    setattr(globals()[module_path], func_name, original_func)

            self._patched = False
            self._original_functions = {}

        except Exception as e:
            raise RuntimeError(f"Error unpatching OpenAI SDK: {str(e)}")

        def _rollback_patches(self) -> None:
            """Rollback patches if patching fails."""
            self.unpatch()

    def _wrap_chat_completion(self, func: Any) -> Any:
        """Wrap sync chat completion with tracing and token tracking."""

        @functools.wraps(func)
        def wrapper(client_self: ChatCompletions, *args: Any,
                    **kwargs: Any) -> Any:
            with start_llm_span("openai.chat.completion", kwargs) as span:
                self.logger.info("Starting chat completion request",
                                 extra={"model": kwargs.get('model')})
                try:
                    messages = kwargs.get('messages', [])
                    model = kwargs.get('model', 'gpt-3.5-turbo')
                    prompt_tokens = count_prompt_tokens(messages, model)
                    span.set_attribute("prompt.tokens", prompt_tokens)
                    prompt = messages[-1]['content'] if messages else ""

                    if kwargs.get('stream', False):
                        return handle_stream(func,
                                             client_self,
                                             span,
                                             prompt_tokens,
                                             self._token_tracker,
                                             self._context,
                                             prompt=prompt,
                                             is_chat=True,
                                             *args,
                                             **kwargs)

                    response = func(client_self, *args, **kwargs)
                    total_tokens = response.usage.total_tokens if hasattr(
                        response, 'usage') else 0
                    completion_tokens = response.usage.completion_tokens if hasattr(
                        response, 'usage') else 0

                    span.set_attribute("completion.tokens", completion_tokens)
                    span.set_attribute("total.tokens", total_tokens)

                    if hasattr(response, 'usage'):
                        update_token_usage(self._token_tracker, "openai",
                                           response.usage)

                    completion = response.choices[
                        0].message.content if response.choices else ""
                    if self._context and self._context.policy_engine:
                        enforce_policies(self._context,
                                         span,
                                         response,
                                         prompt=prompt,
                                         completion=completion)

                    return response

                except Exception as e:
                    span.record_exception(e)
                    raise

        return wrapper

    def _wrap_async_chat_completion(self, func: Any) -> Any:
        """Wrap async chat completion with tracing and token tracking."""

        @functools.wraps(func)
        async def wrapper(client_self: AsyncChatCompletions, *args: Any,
                          **kwargs: Any) -> Any:
            with start_llm_span("openai.chat.completion.async",
                                kwargs) as span:
                self.logger.info("Starting async chat completion request",
                                 extra={"model": kwargs.get('model')})
                try:
                    messages = kwargs.get('messages', [])
                    model = kwargs.get('model', 'gpt-3.5-turbo')
                    prompt_tokens = count_prompt_tokens(messages, model)
                    span.set_attribute("prompt.tokens", prompt_tokens)
                    prompt = messages[-1]['content'] if messages else ""

                    if kwargs.get('stream', False):
                        return await handle_async_stream(func,
                                                         client_self,
                                                         span,
                                                         prompt_tokens,
                                                         self._token_tracker,
                                                         self._context,
                                                         prompt=prompt,
                                                         is_chat=True,
                                                         *args,
                                                         **kwargs)

                    response = await func(client_self, *args, **kwargs)
                    total_tokens = response.usage.total_tokens if hasattr(
                        response, 'usage') else 0
                    completion_tokens = response.usage.completion_tokens if hasattr(
                        response, 'usage') else 0

                    span.set_attribute("completion.tokens", completion_tokens)
                    span.set_attribute("total.tokens", total_tokens)

                    if hasattr(response, 'usage'):
                        update_token_usage(self._token_tracker, "openai",
                                           response.usage)

                    completion = response.choices[
                        0].message.content if response.choices else ""

                    enforce_policies(self._context,
                                     span,
                                     response,
                                     prompt=prompt,
                                     completion=completion)
                    return response

                except Exception as e:
                    span.record_exception(e)
                    raise

        return wrapper

    def _wrap_completion(self, func: Any) -> Any:
        """Wrap sync completion with tracing and token tracking."""

        @functools.wraps(func)
        def wrapper(client_self: Completions, *args: Any,
                    **kwargs: Any) -> Any:
            with start_llm_span("openai.completion", kwargs) as span:
                self.logger.info("Starting completion request",
                                 extra={"model": kwargs.get('model')})
                try:
                    prompt = kwargs.get('prompt', '')
                    model = kwargs.get('model', 'gpt-3.5-turbo')
                    prompt_tokens = count_text_tokens(prompt, model) if isinstance(prompt, str) else \
                                  sum(count_text_tokens(p, model) for p in prompt)

                    span.set_attribute("prompt.tokens", prompt_tokens)

                    if kwargs.get('stream', False):
                        return handle_stream(func,
                                             client_self,
                                             span,
                                             prompt_tokens,
                                             self._token_tracker,
                                             self._context,
                                             prompt=prompt,
                                             is_chat=False,
                                             *args,
                                             **kwargs)

                    response = func(client_self, *args, **kwargs)
                    total_tokens = response.usage.total_tokens if hasattr(
                        response, 'usage') else 0
                    completion_tokens = response.usage.completion_tokens if hasattr(
                        response, 'usage') else 0

                    span.set_attribute("completion.tokens", completion_tokens)
                    span.set_attribute("total.tokens", total_tokens)

                    if hasattr(response, 'usage'):
                        update_token_usage(self._token_tracker, "openai",
                                           response.usage)

                    completion = response.choices[
                        0].text if response.choices else ""
                    enforce_policies(
                        self._context,
                        span,
                        response,
                        prompt=prompt if isinstance(prompt, str) else
                        prompt[0],  # Use first prompt if multiple
                        completion=completion)
                    return response

                except Exception as e:
                    span.record_exception(e)
                    raise

        return wrapper

    def _wrap_async_completion(self, func: Any) -> Any:
        """Wrap async completion with tracing and token tracking."""

        @functools.wraps(func)
        async def wrapper(client_self: AsyncCompletions, *args: Any,
                          **kwargs: Any) -> Any:
            with start_llm_span("openai.completion.async", kwargs) as span:
                self.logger.info("Starting async completion request",
                                 extra={"model": kwargs.get('model')})
                try:
                    prompt = kwargs.get('prompt', '')
                    model = kwargs.get('model', 'gpt-3.5-turbo')
                    prompt_tokens = count_text_tokens(prompt, model) if isinstance(prompt, str) else \
                                  sum(count_text_tokens(p, model) for p in prompt)

                    span.set_attribute("prompt.tokens", prompt_tokens)

                    if kwargs.get('stream', False):
                        return await handle_async_stream(func,
                                                         client_self,
                                                         span,
                                                         prompt_tokens,
                                                         self._token_tracker,
                                                         self._context,
                                                         is_chat=False,
                                                         *args,
                                                         **kwargs)

                    response = await func(client_self, *args, **kwargs)
                    total_tokens = response.usage.total_tokens if hasattr(
                        response, 'usage') else 0
                    completion_tokens = response.usage.completion_tokens if hasattr(
                        response, 'usage') else 0

                    span.set_attribute("completion.tokens", completion_tokens)
                    span.set_attribute("total.tokens", total_tokens)

                    if hasattr(response, 'usage'):
                        update_token_usage(self._token_tracker, "openai",
                                           response.usage)

                    completion = response.choices[
                        0].text if response.choices else ""
                    enforce_policies(
                        self._context,
                        span,
                        response,
                        prompt=prompt if isinstance(prompt, str) else
                        prompt[0],  # Use first prompt if multiple
                        completion=completion)
                    return response

                except Exception as e:
                    span.record_exception(e)
                    raise

        return wrapper

    def _wrap_embedding(self, func: Any) -> Any:
        """Wrap sync embedding with tracing and token tracking."""

        @functools.wraps(func)
        def wrapper(client_self: Embeddings, *args: Any, **kwargs: Any) -> Any:
            with start_llm_span("openai.embeddings", kwargs) as span:
                try:
                    input_text = kwargs.get('input', '')
                    model = kwargs.get('model', 'text-embedding-ada-002')
                    input_tokens = count_text_tokens(input_text, model) if isinstance(input_text, str) else \
                                 sum(count_text_tokens(text, model) for text in input_text)

                    span.set_attribute("input.tokens", input_tokens)
                    response = func(client_self, *args, **kwargs)

                    total_tokens = response.usage.total_tokens if hasattr(
                        response, 'usage') else input_tokens
                    span.set_attribute("total.tokens", total_tokens)

                    if hasattr(response, 'usage'):
                        update_token_usage(self._token_tracker, "openai",
                                           response.usage)
                    return response

                except Exception as e:
                    span.record_exception(e)
                    raise

        return wrapper

    def _wrap_async_embedding(self, func: Any) -> Any:
        """Wrap async embedding with tracing and token tracking."""

        @functools.wraps(func)
        async def wrapper(client_self: AsyncEmbeddings, *args: Any,
                          **kwargs: Any) -> Any:
            with start_llm_span("openai.embeddings.async", kwargs) as span:
                try:
                    input_text = kwargs.get('input', '')
                    model = kwargs.get('model', 'text-embedding-ada-002')
                    input_tokens = count_text_tokens(input_text, model) if isinstance(input_text, str) else \
                                 sum(count_text_tokens(text, model) for text in input_text)

                    span.set_attribute("input.tokens", input_tokens)
                    response = await func(client_self, *args, **kwargs)

                    total_tokens = response.usage.total_tokens if hasattr(
                        response, 'usage') else input_tokens
                    span.set_attribute("total.tokens", total_tokens)

                    if hasattr(response, 'usage'):
                        update_token_usage(self._token_tracker, "openai",
                                           response.usage)
                    return response

                except Exception as e:
                    span.record_exception(e)
                    raise

        return wrapper

    def _wrap_file_upload(self, func: Any) -> Any:
        """Wrap sync file upload with tracing."""

        @functools.wraps(func)
        def wrapper(client_self: Files, *args: Any, **kwargs: Any) -> Any:
            with start_llm_span("openai.files.upload", kwargs) as span:
                try:
                    file = kwargs.get('file')
                    if hasattr(file, 'seek') and hasattr(file, 'tell'):
                        current_pos = file.tell()
                        file.seek(0, os.SEEK_END)
                        file_size = file.tell()
                        file.seek(current_pos)
                        span.set_attribute("file.size_bytes", file_size)

                    return func(client_self, *args, **kwargs)

                except Exception as e:
                    span.record_exception(e)
                    raise

        return wrapper

    def _wrap_async_file_upload(self, func: Any) -> Any:
        """Wrap async file upload with tracing."""

        @functools.wraps(func)
        async def wrapper(client_self: AsyncFiles, *args: Any,
                          **kwargs: Any) -> Any:
            with start_llm_span("openai.files.upload.async", kwargs) as span:
                try:
                    file = kwargs.get('file')
                    if hasattr(file, 'seek') and hasattr(file, 'tell'):
                        current_pos = file.tell()
                        file.seek(0, os.SEEK_END)
                        file_size = file.tell()
                        file.seek(current_pos)
                        span.set_attribute("file.size_bytes", file_size)

                    return await func(client_self, *args, **kwargs)

                except Exception as e:
                    span.record_exception(e)
                    raise

        return wrapper

    def _wrap_image_generation(self, func: Any) -> Any:
        """Wrap sync image generation with tracing."""

        @functools.wraps(func)
        def wrapper(client_self: Images, *args: Any, **kwargs: Any) -> Any:
            with start_llm_span("openai.images.generate", kwargs) as span:
                try:
                    prompt = kwargs.get('prompt', '')
                    prompt_tokens = count_text_tokens(prompt, "gpt-3.5-turbo")

                    span.set_attributes({
                        "prompt.tokens":
                        prompt_tokens,
                        "image.size":
                        kwargs.get('size', 'unknown'),
                        "image.quality":
                        kwargs.get('quality', 'standard'),
                        "image.style":
                        kwargs.get('style', 'natural')
                    })

                    response = func(client_self, *args, **kwargs)
                    span.set_attribute(
                        "images.generated",
                        len(response.data) if hasattr(response, 'data') else 0)
                    return response

                except Exception as e:
                    span.record_exception(e)
                    raise

        return wrapper

    def _wrap_async_image_generation(self, func: Any) -> Any:
        """Wrap async image generation with tracing."""

        @functools.wraps(func)
        async def wrapper(client_self: AsyncImages, *args: Any,
                          **kwargs: Any) -> Any:
            with start_llm_span("openai.images.generate.async",
                                kwargs) as span:
                try:
                    prompt = kwargs.get('prompt', '')
                    prompt_tokens = count_text_tokens(prompt, "gpt-3.5-turbo")

                    span.set_attributes({
                        "prompt.tokens":
                        prompt_tokens,
                        "image.size":
                        kwargs.get('size', 'unknown'),
                        "image.quality":
                        kwargs.get('quality', 'standard'),
                        "image.style":
                        kwargs.get('style', 'natural')
                    })

                    response = await func(client_self, *args, **kwargs)
                    span.set_attribute(
                        "images.generated",
                        len(response.data) if hasattr(response, 'data') else 0)
                    return response

                except Exception as e:
                    span.record_exception(e)
                    raise

        return wrapper

    def __enter__(self) -> 'OpenAIPatcher':
        """Enable use as a context manager."""
        self.patch()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Cleanup when exiting context."""
        self.unpatch(self._original_functions)



from dataclasses import dataclass, field
from typing import Dict, List, Optional, Set
from datetime import datetime
from opentelemetry import trace, baggage
from opentelemetry.trace import Span, SpanKind
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, ConsoleSpanExporter
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter

from .policy_engine import PolicyEngine, PolicyResult, Policy
from ..utils.logging import FileSpanExporter


@dataclass
class TraceContext:
    """Core trace context with essential fields"""
    trace_id: str
    parent_id: Optional[str]
    attributes: Dict
    session_id: Optional[str] = None
    user_id: Optional[str] = None
    active_policies: Set[str] = field(default_factory=set)
    violation_count: int = 0
    policy_results: List[PolicyResult] = field(default_factory=list)

    def to_dict(self) -> Dict:
        """Convert context to dictionary for span attributes"""
        return {
            "trace_id": self.trace_id,
            "parent_id": self.parent_id or "",
            "session_id": self.session_id or "",
            "user_id": self.user_id or "",
            "active_policies": list(self.active_policies),
            "violation_count": self.violation_count,
            **self.attributes
        }

    def set_user_id(self, user_id: str) -> None:
        """Set the user ID for this context"""
        self.user_id = user_id

    def add_violation(self, policy_result: PolicyResult) -> None:
        """Record a policy violation"""
        self.violation_count += len(policy_result.violations)
        self.policy_results.append(policy_result)

    def add_policy(self, policy: str) -> None:
        """Add an active policy"""
        self.active_policies.add(policy)

    def remove_policy(self, policy: str) -> None:
        """Remove an active policy"""
        self.active_policies.discard(policy)


class ContextManager:

    def __init__(self,
                 service_name: str,
                 otel_endpoint: Optional[str] = None,
                 opa_endpoint: Optional[str] = None,
                 policies: Optional[List[Policy]] = None,
                 log_file: Optional[str] = None,
                 trace_console: bool = False):
        self._sessions: Dict[str, TraceContext] = {}
        self._service_name = service_name
        self._current_user_id: Optional[str] = None
        self._log_file = log_file

        # Initialize policy engine if OPA endpoint is provided
        self.policy_engine = PolicyEngine(
            opa_endpoint=opa_endpoint,
            policies=policies) if opa_endpoint else None

        # Set up tracing
        provider = TracerProvider()

        if otel_endpoint:
            otlp_processor = BatchSpanProcessor(
                OTLPSpanExporter(endpoint=otel_endpoint))
            provider.add_span_processor(otlp_processor)

        if trace_console:
            console_processor = BatchSpanProcessor(ConsoleSpanExporter())
            provider.add_span_processor(console_processor)

        if log_file:
            file_processor = BatchSpanProcessor(FileSpanExporter(log_file))
            provider.add_span_processor(file_processor)

        trace.set_tracer_provider(provider)
        self._tracer = trace.get_tracer(service_name)

    def get_session(self, session_id: str) -> Optional[TraceContext]:
        """Get existing session context"""
        return self._sessions.get(session_id)

    def set_user_id(self, user_id: Optional[str]) -> None:
        """Set the global user ID for all new traces"""
        self._current_user_id = user_id

    def get_user_id(self) -> Optional[str]:
        """Get the current global user ID"""
        return self._current_user_id

    async def create_span(
        self,
        name: str,
        context: Optional[TraceContext] = None,
        kind: SpanKind = SpanKind.INTERNAL,
        attributes: Optional[Dict] = None,
    ) -> Span:
        """Create a new span with context"""
        span = self._tracer.start_span(name=name,
                                       kind=kind,
                                       attributes={
                                           "service.name": self._service_name,
                                           **(attributes or {})
                                       })

        if context:
            span.set_attributes(context.to_dict())
            baggage.set_baggage("session_id", context.session_id or "")

        return span

    def create_session(self,
                       session_id: str,
                       initial_context: Optional[Dict] = None) -> TraceContext:
        """Create a new session context"""
        context = TraceContext(trace_id=str(session_id),
                               parent_id=None,
                               attributes=initial_context or {},
                               session_id=session_id,
                               user_id=self._current_user_id)
        self._sessions[session_id] = context
        return context

    async def evaluate_policies(
            self,
            context: TraceContext,
            eval_context: Dict,
            policies: Optional[List[str]] = None) -> PolicyResult:
        """Evaluate policies for the given context"""
        if not self.policy_engine:
            return PolicyResult(passed=True, violations=[])

        policies = policies or list(context.active_policies)
        result = await self.policy_engine.evaluate_with_context(
            eval_context=eval_context, policies=policies)

        if not result.passed:
            context.add_violation(result)

        return result


class ObservabilityContext:
    _instance: Optional[ContextManager] = None

    @classmethod
    def initialize(cls,
                   service_name: str,
                   otel_endpoint: Optional[str] = None,
                   opa_endpoint: Optional[str] = None,
                   policies: Optional[List[Policy]] = None,
                   log_file: Optional[str] = None,
                   trace_console: bool = False) -> None:
        """Initialize the global context manager.
        
        Args:
            service_name: Name of the service using the SDK
            otel_endpoint: OpenTelemetry endpoint for tracing
            opa_endpoint: OPA server endpoint for policy evaluation
            policies: List of Policy objects defining available policies
            log_file: File path for logging output
            trace_console: Whether to enable console tracing
        """
        if cls._instance is None:
            cls._instance = ContextManager(service_name,
                                           otel_endpoint=otel_endpoint,
                                           opa_endpoint=opa_endpoint,
                                           policies=policies,
                                           log_file=log_file,
                                           trace_console=trace_console)

    @classmethod
    def get_current(cls) -> Optional[ContextManager]:
        """Get current context manager instance"""
        if cls._instance is None:
            raise RuntimeError("ObservabilityContext not initialized")
        return cls._instance

    @classmethod
    def create_session(cls,
                       session_id: str,
                       initial_context: Optional[Dict] = None) -> TraceContext:
        """Create a new session context"""
        if cls._instance is None:
            raise RuntimeError("ObservabilityContext not initialized")
        return cls._instance.create_session(session_id, initial_context)

    @classmethod
    def set_user_id(cls, user_id: Optional[str]) -> None:
        """Set the user ID for all new traces"""
        if cls._instance is None:
            raise RuntimeError("ObservabilityContext not initialized")
        cls._instance.set_user_id(user_id)

    @classmethod
    def get_user_id(cls) -> Optional[str]:
        """Get the current user ID"""
        if cls._instance is None:
            raise RuntimeError("ObservabilityContext not initialized")
        return cls._instance.get_user_id()



from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(name="observicia",
      version="0.1.4",
      description="Cloud Native Observability SDK for LLM applications",
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/observicia/observicia",
      packages=find_packages(where="sdk"),
      package_dir={"": "sdk"},
      python_requires=">=3.8",
      install_requires=[
          "numpy>=1.26.0",
          "requests>=2.32.3",
          "tiktoken>=0.7.0",
          "opentelemetry-api>=1.22.0",
          "opentelemetry-sdk>=1.22.0",
          "opentelemetry-exporter-otlp>=1.22.0",
          "opentelemetry-instrumentation>=0.43b0",
          "openai>=1.55.3",
          "pyyaml>=6.0",
      ],
      classifiers=[
          "Development Status :: 4 - Beta",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: OS Independent",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9",
          "Programming Language :: Python :: 3.10",
          "Programming Language :: Python :: 3.11",
          "Topic :: Software Development :: Libraries :: Python Modules",
          "Topic :: System :: Monitoring",
      ],
      keywords="llm, observability, monitoring, tracing, opentelemetry, opa",
      project_urls={
          "Bug Reports": "https://github.com/observicia/observicia/issues",
          "Source": "https://github.com/observicia/observicia",
      })



"""
LLM SDK patchers for various providers.
"""

from .openai import OpenAIPatcher
from .anthropic import AnthropicPatcher
from .litellm import LiteLLMPatcher
from .watsonx import WatsonxPatcher

DEFAULT_PATCHERS = {
    "openai": OpenAIPatcher,
    "anthropic": AnthropicPatcher,
    "litellm": LiteLLMPatcher,
    "watsonx": WatsonxPatcher
}

__all__ = [
    "OpenAIPatcher", "AnthropicPatcher", "LiteLLMPatcher", "WatsonxPatcher",
    "DEFAULT_PATCHERS"
]



"""Utility functions for handling streaming responses"""

from typing import Any, AsyncGenerator, Generator
from opentelemetry import trace
from opentelemetry.trace import Span, get_tracer, SpanKind, Status, StatusCode

from .token_helpers import count_text_tokens
from .policy_helpers import enforce_policies


def _extract_content_from_chunk(chunk: Any, is_chat: bool = False) -> str:
    """Extract content from a response chunk based on type."""
    if not chunk.choices or not chunk.choices[0]:
        return ""

    if is_chat:
        # Chat completion format
        delta = getattr(chunk.choices[0], 'delta', None)
        if delta:
            return delta.content or ""
    else:
        # Regular completion format
        return chunk.choices[0].text or ""

    return ""


async def handle_async_stream(func: Any,
                              client: Any,
                              parent_span: Span,
                              prompt_tokens: int,
                              token_tracker: Any,
                              context: Any,
                              prompt: str = None,
                              is_chat: bool = False,
                              *args: Any,
                              **kwargs: Any) -> AsyncGenerator:
    """Handle async streaming responses."""
    tracer = get_tracer(__name__)
    accumulated_response = []
    chunk_count = 0

    # Get the generator first
    response_generator = await func(client, *args, **kwargs)

    # Create parent context for streaming operation
    parent_ctx = trace.set_span_in_context(parent_span)

    # Create a new span for the entire streaming operation as child of parent span
    with tracer.start_span("stream_processing",
                           context=parent_ctx,
                           kind=SpanKind.INTERNAL) as stream_span:
        stream_span.set_attribute("prompt.tokens", prompt_tokens)
        stream_span.set_attribute("streaming", True)
        if prompt:
            stream_span.set_attribute("has_prompt", True)

        # Create context for child spans
        stream_ctx = trace.set_span_in_context(stream_span)

        async def wrapped_generator():
            nonlocal chunk_count

            try:
                async for chunk in response_generator:
                    content = _extract_content_from_chunk(chunk, is_chat)
                    if content:
                        accumulated_response.append(content)
                        chunk_count += 1
                        # Create chunk span as child of stream span
                        with tracer.start_span(
                                "process_chunk",
                                context=stream_ctx,
                                kind=SpanKind.INTERNAL) as chunk_span:
                            chunk_span.set_attribute("stream.chunks_received",
                                                     chunk_count)
                    yield chunk

                # After stream completes, process accumulated response
                with tracer.start_span("finalize_stream",
                                       context=stream_ctx,
                                       kind=SpanKind.INTERNAL) as final_span:
                    full_response = ''.join(accumulated_response)
                    model = kwargs.get('model', 'gpt-3.5-turbo')
                    completion_tokens = count_text_tokens(full_response, model)
                    total_tokens = prompt_tokens + completion_tokens

                    final_span.set_attribute("completion.tokens",
                                             completion_tokens)
                    final_span.set_attribute("total.tokens", total_tokens)
                    final_span.set_attribute("stream.total_chunks",
                                             chunk_count)

                    token_tracker.update("openai",
                                         prompt_tokens=prompt_tokens,
                                         completion_tokens=completion_tokens)

                    # Structure response based on type
                    response_content = {
                        'choices': [{
                            'message': {
                                'content': full_response
                            }
                        }]
                    } if is_chat else {
                        'choices': [{
                            'text': full_response
                        }]
                    }

                    if context and context.policy_engine:
                        await enforce_policies(context,
                                               final_span,
                                               response_content,
                                               prompt=prompt,
                                               completion=full_response)

            except Exception as e:
                stream_span.record_exception(e)
                raise

        return wrapped_generator()


def handle_stream(func: Any,
                  client: Any,
                  parent_span: Span,
                  prompt_tokens: int,
                  token_tracker: Any,
                  context: Any,
                  prompt: str = None,
                  is_chat: bool = False,
                  *args: Any,
                  **kwargs: Any) -> Generator:
    """Handle sync streaming responses."""
    tracer = get_tracer(__name__)
    accumulated_response = []
    chunk_count = 0

    # Get the sync generator
    response_generator = func(client, *args, **kwargs)

    # Create a new span for the entire streaming operation
    with tracer.start_as_current_span("stream_processing") as stream_span:
        stream_span.set_attribute("prompt.tokens", prompt_tokens)
        stream_span.set_attribute("streaming", True)
        if prompt:
            stream_span.set_attribute("has_prompt", True)

        try:
            for chunk in response_generator:
                content = _extract_content_from_chunk(chunk, is_chat)
                if content:
                    accumulated_response.append(content)
                    chunk_count += 1
                    with tracer.start_as_current_span(
                            "process_chunk") as chunk_span:
                        chunk_span.set_attribute("stream.chunks_received",
                                                 chunk_count)
                yield chunk

            # After stream completes, process accumulated response
            with tracer.start_as_current_span("finalize_stream") as final_span:
                full_response = ''.join(accumulated_response)
                model = kwargs.get('model', 'gpt-3.5-turbo')
                completion_tokens = count_text_tokens(full_response, model)
                total_tokens = prompt_tokens + completion_tokens

                final_span.set_attribute("completion.tokens",
                                         completion_tokens)
                final_span.set_attribute("total.tokens", total_tokens)
                final_span.set_attribute("stream.total_chunks", chunk_count)

                token_tracker.update("openai",
                                     prompt_tokens=prompt_tokens,
                                     completion_tokens=completion_tokens)

                # Structure response based on type
                response_content = {
                    'choices': [{
                        'message': {
                            'content': full_response
                        }
                    }]
                } if is_chat else {
                    'choices': [{
                        'text': full_response
                    }]
                }

                if context and context.policy_engine:
                    enforce_policies(context,
                                     final_span,
                                     response_content,
                                     prompt=prompt,
                                     completion=full_response)

        except Exception as e:
            stream_span.record_exception(e)
            raise



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Set
import requests
from opentelemetry import trace


@dataclass
class Policy:
    """Represents an OPA policy configuration."""
    name: str
    path: str  # Path part of the URL after /v1/data/
    description: Optional[str] = None
    required_trace_level: str = "basic"
    risk_level: str = "low"


@dataclass
class PolicyResult:
    """Result of a policy evaluation."""
    passed: bool
    violations: List[str]
    required_trace_level: str = "basic"  # "basic" or "enhanced"
    metadata: Dict[str, Any] = field(default_factory=dict)
    risk_level: str = "low"  # "low", "medium", "high", "critical"


@dataclass
class Policy:
    """Represents an OPA policy configuration."""
    name: str
    path: str  # Path part of the URL after /v1/data/
    description: Optional[str] = None
    required_trace_level: str = "basic"
    risk_level: str = "low"


class PolicyEngine:
    """
    Policy engine that integrates with OPA for policy evaluation.
    """

    def __init__(self,
                 opa_endpoint: str,
                 policies: Optional[List[Policy]] = None) -> None:
        """
        Initialize PolicyEngine with OPA endpoint and policies.
        
        Args:
            opa_endpoint: Base URL of the OPA server
            policies: List of Policy objects defining available policies
        """
        self.opa_endpoint = opa_endpoint.rstrip('/')

        self.policies: Dict[str, Policy] = {
            p.name: p
            for p in (policies or [])
        }
        self.active_policies: Set[str] = set(self.policies.keys())
        self.tracer = trace.get_tracer(__name__)

    async def evaluate_with_context(
            self,
            eval_context: Dict[str, Any],
            prompt: Optional[str] = None,
            completion: Optional[str] = None,
            policies: Optional[List[str]] = None) -> PolicyResult:
        """
        Evaluate specified policies with given context.
        
        Args:
            eval_context: Context data to evaluate against policies
            prompt: Optional prompt text for prompt compliance check
            completion: Optional completion text for prompt compliance check
            policies: List of policy names to evaluate. If None, use active policies.
        """
        with self.tracer.start_span("policy_evaluation") as span:
            # Use specified policies or fall back to active policies
            policy_names = policies if policies is not None else list(
                self.active_policies)

            # Filter to only existing policies
            policies_to_evaluate = [
                self.policies[name] for name in policy_names
                if name in self.policies
            ]

            # Add prompt and completion to evaluation context if provided
            if prompt is not None and completion is not None:
                eval_context["prompt"] = prompt
                eval_context["completion"] = completion

            result = await self._evaluate_policies(eval_context,
                                                   policies_to_evaluate)

            span.set_attributes({
                "policy_count":
                len(policies_to_evaluate),
                "violations":
                len(result.violations),
                "risk_level":
                result.risk_level,
                "evaluated_policies":
                ";".join(p.name for p in policies_to_evaluate)
            })

            return result

    def evaluate_sync(self,
                      eval_context: Dict[str, Any],
                      prompt: Optional[str] = None,
                      completion: Optional[str] = None) -> PolicyResult:
        """
        Synchronously evaluate policies with given context.
        """
        with self.tracer.start_span("policy_evaluation") as span:
            policies_to_evaluate = [
                self.policies[name] for name in self.active_policies
                if name in self.policies
            ]

            # Add prompt and completion to evaluation context if provided
            if prompt is not None and completion is not None:
                eval_context["prompt"] = prompt
                eval_context["completion"] = completion

            result = self._evaluate_policies_sync(eval_context,
                                                  policies_to_evaluate)

            span.set_attributes({
                "policy_count":
                len(policies_to_evaluate),
                "violations":
                len(result.violations),
                "risk_level":
                result.risk_level,
                "evaluated_policies":
                ";".join(p.name for p in policies_to_evaluate)
            })

            return result

    def _evaluate_policies_sync(self, eval_context: Dict[str, Any],
                                policies: List[Policy]) -> PolicyResult:
        """Evaluate multiple policies synchronously and aggregate results."""
        violations = []
        metadata = {}
        max_trace_level = "basic"
        max_risk_level = "low"

        for policy in policies:
            raw_result = self._evaluate_single_policy_sync(
                policy, eval_context)
            # print(f'Policy: {policy.name}, Result: {raw_result}')

            # Extract the nested result
            result = raw_result.get('result', {})

            if not result.get("allow", False):
                # Get violations from the nested result
                policy_violations = result.get("violations", [])
                if policy_violations:
                    violations.extend(f"{policy.name}: {v}"
                                      for v in policy_violations)
                else:
                    # If no specific violations but policy failed, add metadata as context
                    policy_metadata = result.get("metadata", {})
                    violation_context = []
                    if policy_metadata.get("pii_detected"):
                        pii_entities = policy_metadata.get("pii_entities", [])
                        for entity in pii_entities:
                            violation_context.append(
                                f"Found {entity['entity_type']} with confidence {entity['score']}"
                            )
                    if violation_context:
                        violations.extend(f"{policy.name}: {v}"
                                          for v in violation_context)
                    else:
                        violations.append(
                            f"{policy.name}: Policy check failed")

            # Get trace level from nested result or policy default
            policy_trace_level = result.get("trace_level",
                                            policy.required_trace_level)
            if policy_trace_level == "enhanced":
                max_trace_level = "enhanced"

            # Get risk level from nested result or policy default
            risk_level = result.get("risk_level", policy.risk_level)
            if self._compare_risk_levels(risk_level, max_risk_level) > 0:
                max_risk_level = risk_level

            # Store metadata from nested result
            metadata[policy.name] = result.get("metadata", {})

        return PolicyResult(passed=len(violations) == 0,
                            violations=violations,
                            required_trace_level=max_trace_level,
                            metadata=metadata,
                            risk_level=max_risk_level)

    def _evaluate_single_policy_sync(
            self, policy: Policy, eval_context: Dict[str,
                                                     Any]) -> Dict[str, Any]:
        """Evaluate a single policy synchronously using OPA."""
        url = f"{self.opa_endpoint}/v1/data/{policy.path}"
        try:
            response = requests.post(url, json={"input": eval_context})
            if response.status_code == 200:
                return response.json()
            else:
                return {
                    "allow":
                    False,
                    "violations":
                    [f"Policy evaluation failed: HTTP {response.status_code}"],
                    "risk_level":
                    "high"
                }
        except Exception as e:
            return {
                "allow": False,
                "violations": [f"Policy evaluation error: {str(e)}"],
                "risk_level": "high"
            }

    def _compare_risk_levels(self, level1: str, level2: str) -> int:
        """Compare two risk levels. Returns positive if level1 > level2."""
        risk_order = {"low": 0, "medium": 1, "high": 2, "critical": 3}
        return risk_order.get(level1, 0) - risk_order.get(level2, 0)

    def add_policy(self, policy: Policy) -> None:
        """Add a policy to the available policies."""
        self.policies[policy.name] = policy
        self.active_policies.add(policy.name)

    def remove_policy(self, policy_name: str) -> None:
        """Remove a policy from the available and active policies."""
        if policy_name in self.policies:
            del self.policies[policy_name]
            self.active_policies.discard(policy_name)

    def activate_policy(self, policy_name: str) -> None:
        """Activate a policy if it exists."""
        if policy_name in self.policies:
            self.active_policies.add(policy_name)

    def deactivate_policy(self, policy_name: str) -> None:
        """Deactivate a policy without removing it."""
        self.active_policies.discard(policy_name)

    @property
    def available_policies(self) -> List[Policy]:
        """Get list of all available policies."""
        return list(self.policies.values())
